#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains functions related with Qt dock behaviour
"""

from __future__ import print_function, division, absolute_import

import logging

from Qt.QtCore import Qt, Signal, QPoint, QRect, QPropertyAnimation, QParallelAnimationGroup, QAbstractAnimation
from Qt.QtWidgets import QApplication, QSizePolicy, QBoxLayout, QStackedLayout, QWidget, QFrame, QSplitter, QLabel
from Qt.QtWidgets import QPushButton, QScrollArea, QGraphicsDropShadowEffect, QStyle, QMenu, QAction
from Qt.QtGui import QCursor, QIcon

from tpDcc.libs.qt.widgets import layouts, drop

LOGGER = logging.getLogger('tpDcc-libs-qt')

USE_ANIMATIONS = True


# ===================================================================================================================

def new_dock_splitter(orientation=Qt.Horizontal, parent=None):
    s = QSplitter(orientation, parent)
    s.setProperty('dock-splitter', True)
    s.setChildrenCollapsible(False)
    s.setOpaqueResize(False)

    return s


def splitter_contains_section_widget(splitter):
    for i in range(splitter.count()):
        w = splitter.widget(i)
        if splitter and splitter_contains_section_widget(splitter):
            return True
        else:
            if issubclass(w, DockSectionWidget):
                return True

    return False


def delete_empty_splitter(container):
    do_again = False
    while do_again:
        do_again = False
        splitters = container.findChildren(QSplitter)
        for i in range(len(splitters)):
            sp = splitters.at(i)
            if not sp.property('dock-splitter').toBool():
                continue
            if sp.countT() > 0 and splitter_contains_section_widget(sp):
                continue

            del splitters[i]
            do_again = True
            break


def find_parent_container_widget(widget):
    next_widget = widget
    while next_widget:
        if isinstance(next_widget, DockContainer):
            break
        next_widget = next_widget.parentWidget()

    return next_widget


def find_parent_section_widget(widget):
    next_widget = widget
    while next_widget:
        if isinstance(next_widget, DockSectionWidget):
            break
        next_widget = next_widget.parentWidget()

    return next_widget


def find_parent_splitter(widget):
    next_widget = widget
    while next_widget:
        if isinstance(next_widget, QSplitter):
            break
        next_widget = next_widget.parentWidget()

    return next_widget


def find_inmediate_splitter(widget):
    sp = None
    layout = widget.layout()
    if not layout or layout.count() <= 0:
        return sp

    for i in range(layout.count()):
        layout_item = layout.itemAt(0)
        if not layout_item.widget():
            continue
        if isinstance(layout_item.widget(), QSplitter):
            sp = layout_item.widget()
            break

    return sp


# ===================================================================================================================

class DockInternalContentData(object):
    content = None
    title_widget = None
    content_widget = None


class DockHiddenSectionItem(object):
    def __init__(self):
        self.preferred_section_id = -1
        self.preferred_section_index = -1
        self.data = DockInternalContentData()


class DockFlags(object):
    NoFlag = 0
    Closeable = 1
    AllFlags = Closeable


class DockSectionContent(object):

    next_uid = 0

    def __init__(self):
        super(DockSectionContent, self).__init__()

        self._uid = self.get_next_uid()
        self._unique_name = ''
        self._title = ''
        self._flags = [DockFlags.AllFlags]
        self._container_widget = None
        self._content_widget = None
        self._title_widget = None

        LOGGER.debug('Creating DockSectionContent {}'.format(self._uid))

    def get_next_uid(self):
        next_id = DockSectionContent.next_uid
        DockSectionContent.next_uid += 1
        return next_id

    def uid(self):
        return self._uid

    def unique_name(self):
        return self._unique_name

    def container_widget(self):
        return self._container_widget

    def content_widget(self):
        return self._content_widget

    def title_widget(self):
        return self._title_widget

    def flags(self):
        return self._flags

    def title(self):
        return self._title

    def set_title(self, title):
        self._title = title

    def set_flags(self, flags):
        self._flags = flags

    def clean(self):
        if self._container_widget:
            del self._container_widget._section_container_lookup_map_by_id[self._uid]
            del self._container_widget._section_container_lookup_map_by_name[self._unique_name]
            self._title_widget.deleteLater()
            del self._title_widget
            self._content_widget.deleteLater()
            del self._content_widget

    @staticmethod
    def new_section_content(unique_name, container, title, content):
        """
        Creates a new DockSectionWidget an register it into its DockContainer
        :param unique_name: str, unique name for the DockSectionWidget
        :param container: DockContainer, container where DockSectionWidget is stored
        :param title: str, title of the DockSectionWidget
        :param content: QWidget, widget that DockSectionWidget should store and manage
        :return: QDockSectionWidget
        """

        if unique_name == '':
            LOGGER.error('Cannot create DockSectionContent with empty unique_name')
            return None
        elif DockContainer.check_section_container_lookup_map_by_name(container, unique_name):
            LOGGER.error('Cannot create DockSectionContent with already used unique_name')
            return None
        elif not container or not title or not content:
            LOGGER.error('Cannot create DockSectionContent with None values')
            return None

        section_content = DockSectionContent()
        section_content._unique_name = unique_name
        section_content._container_widget = container
        section_content._title_widget = QLabel(title)
        section_content._content_widget = content
        container.register_section_content(section_content)

        return section_content

    def visible_title(self):
        if self._title == '':
            return self._unique_name

        return self._title


class DockSectionContentWidget(QFrame, object):
    def __init__(self, section_content, parent=None):
        super(DockSectionContentWidget, self).__init__(parent=parent)

        self._content = section_content

        self.main_layout = QBoxLayout(QBoxLayout.TopToBottom)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        self.main_layout.addWidget(self._content.content_widget())
        self.setLayout(self.main_layout)

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def clean(self):
        self.main_layout.removeWidget(self._content.content_widget())


class DockSectionTitleWidget(QFrame, object):

    clicked = Signal()
    activeTabChanged = Signal(bool)

    def __init__(self, section_content, parent=None):
        super(DockSectionTitleWidget, self).__init__(parent=parent)

        self._content = section_content
        self._tab_moving = False
        self._active_tab = False
        self._drag_start_pos = QPoint()
        self._floating_widget = None

        self.main_layout = QBoxLayout(QBoxLayout.LeftToRight)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        self.main_layout.addWidget(section_content.title_widget())
        self.setLayout(self.main_layout)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            event.accept()
            self._drag_start_pos = event.pos()
            LOGGER.debug('Starting to drag Dock from pos: {}'.format(self._drag_start_pos))
            return

        super(DockSectionTitleWidget, self).mousePressEvent(event)

    def _on_anim_move(self, data, container_widget, section_widget=None, loc=None):
        if section_widget:
            move_anim = QPropertyAnimation(self._floating_widget, 'pos', self)
            move_anim.setStartValue(self._floating_widget.pos())
            move_anim.setEndValue(section_widget.mapToGlobal(section_widget.rect().topLeft()))
            move_anim.setDuration(150)
            resize_anim = QPropertyAnimation(self._floating_widget, 'size', self)
            resize_anim.setStartValue(self._floating_widget.size())
            resize_anim.setEndValue(section_widget.size())
            resize_anim.setDuration(150)
            anim_grp = QParallelAnimationGroup(self)
            anim_grp.finished.connect(lambda: self._on_move_widget(data, container_widget, section_widget, loc))
            anim_grp.addAnimation(move_anim)
            anim_grp.addAnimation(resize_anim)
            anim_grp.start(QAbstractAnimation.DeleteWhenStopped)
        else:
            self._on_move_widget(data, container_widget, section_widget, loc)

    def _on_move_widget(self, data, container_widget, section_widget=None, loc=None):
        data = self._floating_widget.take_content(data)
        self._floating_widget.deleteLater()
        self._floating_widget.clean()
        self._floating_widget = None
        if loc and section_widget:
            container_widget.drop_content(data, section_widget, loc, True)

    def mouseReleaseEvent(self, event):
        container_widget = find_parent_container_widget(widget=self)

        # If we are dragging a widget
        if self._floating_widget:

            # We detect if we are over a DockContentWidget
            section_widget = container_widget.section_at(container_widget.mapFromGlobal(event.globalPos()))
            if section_widget:
                container_widget.drop_overlay().set_allowed_areas(drop.DropArea.AllAreas)
                loc = container_widget.drop_overlay().show_drop_overlay(section_widget)

                # If the drop is in an invalid area we put by default the container in the center
                if loc == drop.DropArea.InvalidDropArea:
                    loc = drop.DropArea.CenterDropArea

                if loc != drop.DropArea.InvalidDropArea:
                    data = DockInternalContentData()
                    if USE_ANIMATIONS:
                        self._on_anim_move(data, container_widget, section_widget, loc)
                    else:
                        self._on_move_widget(data, container_widget, section_widget, loc)
            else:
                drop_area = drop.DropArea.InvalidDropArea
                if container_widget.outer_top_drop_rect().contains(container_widget.mapFromGlobal(event.globalPos())):
                    drop_area = drop.DropArea.TopDropArea
                if container_widget.outer_right_drop_rect().contains(container_widget.mapFromGlobal(event.globalPos())):
                    drop_area = drop.DropArea.RightDropArea
                if container_widget.outer_bottom_drop_rect().contains(
                        container_widget.mapFromGlobal(event.globalPos())):
                    drop_area = drop.DropArea.BottomDropArea
                if container_widget.outer_left_drop_rect().contains(container_widget.mapFromGlobal(event.globalPos())):
                    drop_area = drop.DropArea.LeftDropArea

                # If the drop is in an invalid area we put by default the container in the center
                if drop_area == drop.DropArea.InvalidDropArea:
                    drop_area = drop.DropArea.CenterDropArea

                if drop_area != drop.DropArea.InvalidDropArea:
                    data = DockInternalContentData()
                    if USE_ANIMATIONS:
                        self._on_anim_move(data, container_widget)
                    else:
                        self._on_move_widget(data, container_widget)

                    container_widget.drop_content(data, None, drop_area, True)
        elif self._tab_moving and find_parent_section_widget(self) is not None:
            section = find_parent_section_widget(widget=self)
            pos = event.globalPos()
            pos = section.mapFromGlobal(pos)
            from_index = section.index_of_content(self._content)
            to_index = section.index_of_content_by_title_pos(pos, self)
            section.move_content(from_index, to_index)

        if not self._drag_start_pos.isNull():
            self.clicked.emit()

        self._drag_start_pos = QPoint()
        self._tab_moving = False
        container_widget.drop_overlay().hide_drop_overlay()
        super(DockSectionTitleWidget, self).mouseReleaseEvent(event)

    def mouseMoveEvent(self, event):

        # Get the DockContainer where this title is stored
        container_widget = find_parent_container_widget(widget=self)
        if container_widget is None:
            LOGGER.warning('Title is not stored in DockContainer! This should never happen!')
            return

        # Get the DockSectionWidget where this title is stored
        section_widget = find_parent_section_widget(widget=self)

        if self._floating_widget and (event.buttons() & Qt.LeftButton):
            event.accept()
            move_to_pos = event.globalPos() - (self._drag_start_pos + QPoint(7, 7))
            self._floating_widget.move(move_to_pos)

            # Mouse is over a section widget
            section = container_widget.section_at(container_widget.mapFromGlobal(QCursor.pos()))
            if section:
                container_widget.drop_overlay().set_allowed_areas(drop.DropArea.AllAreas)
                container_widget.drop_overlay().show_drop_overlay(section)

            # Mouse is at the edge of the Container Widget
            elif container_widget.outer_top_drop_rect().contains(container_widget.mapFromGlobal(QCursor.pos())):
                container_widget.drop_overlay().set_allowed_areas([drop.DropArea.TopDropArea])
                container_widget.drop_overlay().show_drop_overlay(
                    container_widget, container_widget.outer_top_drop_rect())
            elif container_widget.outer_right_drop_rect().contains(container_widget.mapFromGlobal(QCursor.pos())):
                container_widget.drop_overlay().set_allowed_areas([drop.DropArea.RightDropArea])
                container_widget.drop_overlay().show_drop_overlay(
                    container_widget, container_widget.outer_right_drop_rect())
            elif container_widget.outer_bottom_drop_rect().contains(container_widget.mapFromGlobal(QCursor.pos())):
                container_widget.drop_overlay().set_allowed_areas([drop.DropArea.BottomDropArea])
                container_widget.drop_overlay().show_drop_overlay(
                    container_widget, container_widget.outer_bottom_drop_rect())
            elif container_widget.outer_left_drop_rect().contains(container_widget.mapFromGlobal(QCursor.pos())):
                container_widget.drop_overlay().set_allowed_areas([drop.DropArea.LeftDropArea])
                container_widget.drop_overlay().show_drop_overlay(
                    container_widget, container_widget.outer_left_drop_rect())
            else:
                container_widget.drop_overlay().hide_drop_overlay()
            return
        elif not self._floating_widget and not self._drag_start_pos.isNull() and (
                event.buttons() & Qt.LeftButton) and section_widget is not None and \
                not section_widget.title_area_geometry().contains(section_widget.mapFromGlobal(event.globalPos())):
            event.accept()

            # We get the new dock info for the current section
            # DockSectionTitleWidget and DockSectionContentWidgets are reparented to the DockContainer and are ready
            # to be moved to a new DockSectionWidget
            LOGGER.debug('Creating new floating dock data ...')
            data = section_widget.take_content(self._content.uid())
            if data is None:
                LOGGER.error(
                    'This should not happen! {} - {}'.format(self._content.uid(), self._content.unique_name()))
                return

            # Create floating widget and add it to the list of Container floatters
            LOGGER.debug('Creating Floating Widget ...')
            self._floating_widget = DockFloatingWidget(container=container_widget, section_content=data.content,
                                                       section_title_widget=data.title_widget,
                                                       content_widget=data.content_widget, parent=container_widget)
            self._floating_widget.resize(section_widget.size())
            container_widget.floatings().append(self._floating_widget)

            move_to_pos = event.globalPos() - (self._drag_start_pos + QPoint(7, 7))
            self._floating_widget.move(move_to_pos)
            self._floating_widget.show()

            # Delete old section, if it is empty now
            if section_widget.is_empty():
                section_widget.clean()
                section_widget.deleteLater()
                del section_widget

            delete_empty_splitter(container=container_widget)
            return
        elif self._tab_moving and find_parent_section_widget(self):
            event.accept()
            move_to_pos = self.mapToParent(event.pos()) - self._drag_start_pos
            move_to_pos.setY(0)
            self.move(move_to_pos)
            return
        elif not self._drag_start_pos.isNull() and (
                event.buttons() & Qt.LeftButton) and (event.pos() - self._drag_start_pos).manhattanLength() >= \
                QApplication.startDragDistance() and find_parent_section_widget(self) is not None \
                and find_parent_section_widget(self).title_area_geometry().contains(
                find_parent_section_widget(self).mapFromGlobal(event.globalPos())):
            event.accept()
            self._tab_moving = True
            self.raise_()
            return

        super(DockSectionTitleWidget, self).mouseMoveEvent(event)

    def clean(self):
        self.main_layout.removeWidget(self._content.title_widget())

    def content(self):
        return self._content

    def is_active_tab(self):
        return self._active_tab

    def set_active_tab(self, active):
        if active != self._active_tab:
            self._active_tab = active
            self.style().unpolish(self)
            self.style().polish(self)
            self.update()
            self.activeTabChanged.emit(active)

    def _on_anim_finished(self, content_widget, data, section_widget, loc):
        data = self._floating_widget.take_content(data)
        self._floating_widget.clean()
        self._floating_widget.deleteLater()
        self._floating_widget = None
        content_widget.drop_content(data, section_widget, loc)


class DockSectionWidgetTabsScrollArea(QScrollArea, object):
    """
    Custom scrollable implementation for tabs
    """

    def __init__(self, section_widget, parent=None):
        super(DockSectionWidgetTabsScrollArea, self).__init__(parent=parent)

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setFrameStyle(QFrame.NoFrame)
        self.setWidgetResizable(True)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self.setMaximumHeight(15)

    def wheelEvent(self, event):
        event.accept()
        try:
            # For Qt >= 5.0.0
            direction = event.angleDelta().y()
        except Exception:
            direction = event.delta()

        if direction < 0:
            self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() + 20)
        else:
            self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() - 20)


class DockSectionWidget(QFrame, object):

    next_uid = 0

    def __init__(self, container_widget=None):
        super(DockSectionWidget, self).__init__(parent=container_widget)

        self._uid = self.get_next_uid()
        self._container = container_widget
        self._tabs_layout_init_count = 0
        self._mouse_press_title_widget = None

        self._contents = list()
        self._section_titles = list()
        self._section_contents = list()

        self.main_layout = QBoxLayout(QBoxLayout.TopToBottom)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        self.setLayout(self.main_layout)

        self._top_layout = QBoxLayout(QBoxLayout.LeftToRight)
        self._top_layout.setContentsMargins(0, 0, 0, 0)
        self._top_layout.setSpacing(0)
        self.main_layout.addLayout(self._top_layout)

        self._tabs_scroll_area = DockSectionWidgetTabsScrollArea(section_widget=self)
        self._top_layout.addWidget(self._tabs_scroll_area, 1)

        self._tabs_container_widget = QWidget()
        self._tabs_container_widget.setObjectName('tabsContainerWidget')
        self._tabs_scroll_area.setWidget(self._tabs_container_widget)

        self._tabs_layout = QBoxLayout(QBoxLayout.LeftToRight)
        self._tabs_layout.setContentsMargins(0, 0, 0, 0)
        self._tabs_layout.setSpacing(0)
        self._tabs_layout.addStretch(1)
        self._tabs_container_widget.setLayout(self._tabs_layout)

        self._tabs_menu_button = QPushButton()
        self._tabs_menu_button.setObjectName('tabsMenuButton')
        self._tabs_menu_button.setFlat(True)
        self._tabs_menu_button.setIcon(self.style().standardIcon(QStyle.SP_TitleBarUnshadeButton))
        self._tabs_menu_button.setMaximumWidth(self._tabs_menu_button.iconSize().width())
        self._top_layout.addWidget(self._tabs_menu_button, 0)

        self._close_btn = QPushButton()
        self._close_btn.setObjectName('closeButton')
        self._close_btn.setFlat(True)
        self._close_btn.setIcon(self.style().standardIcon(QStyle.SP_TitleBarCloseButton))
        self._close_btn.setMaximumWidth(self._close_btn.iconSize().width())
        self._top_layout.addWidget(self._close_btn, 0)
        self._close_btn.clicked.connect(self._on_close_button_clicked)

        self._tabs_layout_init_count = self._tabs_layout.count()

        # Central area with contents
        self._contents_layout = QStackedLayout()
        self._contents_layout.setContentsMargins(0, 0, 0, 0)
        self._contents_layout.setSpacing(0)
        self.main_layout.addLayout(self._contents_layout, 1)

        shadow_effect = QGraphicsDropShadowEffect(self)
        shadow_effect.setOffset(0, 0)
        shadow_effect.setBlurRadius(8)
        self.setGraphicsEffect(shadow_effect)

    @staticmethod
    def get_next_uid():
        next_id = DockSectionWidget.next_uid
        DockSectionWidget.next_uid += 1
        return next_id

    def uid(self):
        return self._uid

    def container_widget(self):
        return self._container

    def contents(self):
        return self._contents

    def is_empty(self):
        return len(self._contents) <= 0

    def showEvent(self, event):
        """
        When the DockSectionWidget is showed, we make need to make sure that the widget is visible
        :param event: QShowEvent
        """

        try:
            self._tabs_scroll_area.ensureWidgetVisible(self._section_titles[self.current_index()])
        except Exception:
            pass

    def title_area_geometry(self):
        return self._top_layout.geometry()

    def content_area_geometry(self):
        return self._contents_layout.geometry()

    def clean(self):
        if self._container:
            del self._container._section_widget_lookup_map_by_id[self._uid]
            self._container._sections.remove(self)

        splitter = find_parent_splitter(self)
        if splitter and splitter.count() == 0:
            splitter.deleteLater()
            del splitter

    def current_index(self):
        """
        Returns the widget index that QStackedWidget is showing right now
        :return: int
        """

        return self._contents_layout.currentIndex()

    def set_current_index(self, index):
        if index < 0 or index > len(self._contents) - 1:
            LOGGER.warning('Invalid index: {}'.format(index))
            return

        for i in range(self._tabs_layout.count()):
            item = self._tabs_layout.itemAt(i)
            if item.widget():
                section_title_widet = item.widget()
                if isinstance(section_title_widet, DockSectionTitleWidget):
                    if i == index:
                        section_title_widet.set_active_tab(True)
                        self._tabs_scroll_area.ensureWidgetVisible(section_title_widet)
                        if DockFlags.Closeable in section_title_widet._content.flags():
                            self._close_btn.setEnabled(True)
                        else:
                            self._close_btn.setEnabled(False)
                    else:
                        section_title_widet.set_active_tab(False)

        # Set active content
        self._contents_layout.setCurrentIndex(index)

    def update_tabs_menu(self):
        m = QMenu()
        for i in range(len(self._contents)):
            section_content = self._contents[i]
            a = m.addAction(QIcon(), section_content.visible_title())
            a.setData(section_content.uid())
            a.triggered.connect(self._on_tabs_menu_action_triggered)

        old = self._tabs_menu_button.menu()
        self._tabs_menu_button.setMenu(m)

        if old:
            old.deleteLater()
            del old

    def add_content(self, section_content):
        """
        Creates DockSectionTitleWidget and DockSectionContentWidget from scratch and store on them
        the given DockSectionContent
        :param section_content: DockSectionContent
        """

        # Add the given DockSectionContent to the list of contents
        self._contents.append(section_content)

        # Create new DockSectionTitle and add it to the list of title widgets and to the layout
        title = DockSectionTitleWidget(section_content=section_content)
        self._section_titles.append(title)
        self._tabs_layout.insertWidget(self._tabs_layout.count() - self._tabs_layout_init_count, title)
        title.clicked.connect(self._on_section_title_clicked)

        # Create DockSectionContentWidget that will store the given DockSectionContent
        content = DockSectionContentWidget(section_content=section_content)
        self._section_contents.append(content)
        self._contents_layout.addWidget(content)

        # If we only have stored one DockSectionContent, we set the current index to the first one
        if len(self._contents) == 1:
            self.set_current_index(0)

        # Update the tab menu depending on the DockSectionContents number
        self.update_tabs_menu()

    def add_data_content(self, internal_data, auto_activate):
        """
        Adds DockSectionTitleWidget and DockSectionContentWidget from the given data. Also the DockSectionContent is
        retrieved from the data
        :param internal_data: DockInternalContentData
        :param auto_activate: bool
        """

        # Add the DockSectionContent stored in the data into the list of contents
        self._contents.append(internal_data.content)

        # Add the DockSectionTitle stored in the data to the list of title widgets and to the layout
        self._section_titles.append(internal_data.title_widget)
        self._tabs_layout.insertWidget(
            self._tabs_layout.count() - self._tabs_layout_init_count, internal_data.title_widget)
        internal_data.title_widget.clicked.connect(self._on_section_title_clicked)

        # Add DockSectionContentWidget stored in the data and associates the DockSectionContent stored also in the data
        self._section_contents.append(internal_data.content_widget)
        self._contents_layout.addWidget(internal_data.content_widget)

        if len(self._contents) == 1:
            self.set_current_index(0)
        elif auto_activate:
            self.set_current_index(len(self._contents) - 1)
        else:
            internal_data.title_widget.set_active_tab(False)

        # Update the tab menu depending on the DockSectionContents number
        self.update_tabs_menu()

    def take_content(self, uid):
        """

        :param uid: uid of the DockSectionContent we want to extract from this section
        :return: bool, whether the new data is valid or not
        """

        section_content = None
        index = -1

        # Check if the given UID is valid and is already stored in this DockSectionWidget
        # Once, is found, we store its index and we extract it from the list of DockSectionContents
        # If no valid DockSectionContent is found, we abort the process
        for i in range(len(self._contents)):
            if self._contents[i].uid() != uid:
                continue
            index = i
            section_content = self._contents.pop(i)
            break
        if section_content is None:
            return None

        # We get the DockSectionTitleWidget associated with the DockSectionContent and:
        # 1. Remove it from the DockSectionWidget layout
        # 2. Disconnect all its signals
        # 3. Parent it to the DockContainer
        title = self._section_titles.pop(index)
        if title:
            self._tabs_layout.removeWidget(title)
            # title.setAttribute(Qt.WA_WState_Created, False)
            title.disconnect(self)
            title.setParent(self._container)
            # title.setAttribute(Qt.WA_WState_Created, True)

        # We get the DockSectionContentWidget associated with the DockSectionContent and:
        # 1. Remove it from the DockSectionWidget
        # 2. Disconnect all its signals
        # 3. Parent it to the DockContainer
        content = self._section_contents.pop(index)
        if content:
            self._contents_layout.removeWidget(content)
            content.disconnect(self)
            content.setParent(self._container)

        # We update the active tab
        if len(self._contents) > 0 and title.is_active_tab():
            if index > 0:
                self.set_current_index(index - 1)
            else:
                self.set_current_index(0)

        # We update the tabs menu
        self.update_tabs_menu()

        if section_content is None or title is None or content is None:
            return None

        internal_data = DockInternalContentData()
        internal_data.content = section_content
        internal_data.title_widget = title
        internal_data.content_widget = content

        # try:
        #     return not internal_data.content.isNull()
        # except Exception:
        #     return internal_data.content is not None
        return internal_data

    def index_of_content(self, section_content):
        if section_content in self._contents:
            return self._contents.index(section_content)

    def index_of_content_by_uid(self, uid):
        for i in range(len(self._contents)):
            if self._contents[i].uid() == uid:
                return i

        return -1

    def index_of_content_by_title_pos(self, p, exclude):
        index = -1
        for i in range(len(self._section_titles)):
            if self._section_titles[i].geometry().contains(p) \
                    and (exclude is None or self._section_titles[i] != exclude):
                index = i
                break

        return index

    def move_content(self, from_index, to_index):
        if from_index >= len(self._contents) or from_index < 0 or \
                to_index >= len(self._contents) or to_index < 0 or from_index == to_index:
            LOGGER.warning('Invalid for tab movement - From: {} | To: {}'.format(from_index, to_index))
            self._tabs_layout.update()
            return

        self._contents.move(from_index, to_index)
        self._section_titles.move(from_index, to_index)
        self._section_contents.move(from_index, to_index)

        layout_from = self._tabs_layout.takeAt(from_index)
        self._tabs_layout.insertWidget(to_index, layout_from.widget())
        layout_from.deleteLater()
        del layout_from
        layout_from = None

        layout_from = self._contents_layout.takeAt(from_index)
        self._contents_layout.insertWidget(to_index, layout_from.widget())
        layout_from.deleteLater()
        del layout_from

        self.update_tabs_menu()

    def _on_close_button_clicked(self):
        index = self.current_index()
        if index < 0 or index > len(self._contents) - 1:
            return
        section_content = self._contents[index]
        if section_content is None:
            return
        self._container.remove_section_content(section_content)

    def _on_section_title_clicked(self):
        section_title_widget = self.sender()
        if not section_title_widget or not isinstance(section_title_widget, DockSectionTitleWidget):
            return

        index = self._tabs_layout.indexOf(section_title_widget)
        self.set_current_index(index)

    def _on_tabs_menu_action_triggered(self, flag):
        a = self.sender()
        if not a or not isinstance(a, QAction):
            return

        uid = a.data().toInt()
        index = self.index_of_content_by_uid(uid=uid)
        if index >= 0:
            self.set_current_index(index)


class DockFloatingWidget(QWidget, object):
    """
    Holds and displays DockSectionContent as a floating window
    It can be resized, moved and dropped back into a DockSectionWidget
    """

    def __init__(self, container, section_content, section_title_widget, content_widget, parent=None):
        super(DockFloatingWidget, self).__init__(parent=parent)

        self._container = container
        self._content = section_content
        self._title_widget = section_title_widget
        self._content_widget = content_widget

        self.main_layout = QBoxLayout(QBoxLayout.TopToBottom)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        self.setLayout(self.main_layout)

        self._title_layout = QBoxLayout(QBoxLayout.LeftToRight)
        self._title_layout.addWidget(section_title_widget, 1)
        self.main_layout.addLayout(self._title_layout, 0)
        section_title_widget.set_active_tab(False)

        if DockFlags.Closeable in section_content.flags():
            close_btn = QPushButton()
            close_btn.setObjectName('CloseButton')
            close_btn.setFlat(True)
            close_btn.setIcon(self.style().standardIcon(QStyle.SP_TitleBarCloseButton))
            close_btn.setToolTip('Close')
            close_btn.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
            self._title_layout.addWidget(close_btn)
            close_btn.clicked.connect(self._on_close_button_clicked)

        self.main_layout.addWidget(content_widget, 1)
        content_widget.show()

    def clean(self):
        del self._container._floatings[:]         # Python 2

    def take_content(self, data):
        data.content = self._content
        data.title_widget = self._title_widget
        data.content_widget = self._content_widget
        self._title_layout.removeWidget(self._title_widget)
        self._title_widget.setParent(self._container)
        self._title_widget = None
        self.main_layout.removeWidget(self._content_widget)
        self._content_widget.setParent(self._container)
        self._content_widget = None

        return data

    def _on_close_button_clicked(self):
        self._container.hide_section_content(self._content)


class DockContainer(QFrame, object):

    orientationChanged = Signal()
    activeTabChanged = Signal(object, bool)
    sectionContentVisibilityChanged = Signal(object, bool)

    def __init__(self, parent=None):
        super(DockContainer, self).__init__(parent=parent)

        self._orientation = Qt.Horizontal
        self._drop_overlay = drop.DropOverlay(parent=self)
        self._splitter = None
        self._sections = list()
        self._floatings = list()
        self._hidden_section_contents = dict()

        self._section_container_lookup_map_by_id = dict()
        self._section_container_lookup_map_by_name = dict()
        self._section_widget_lookup_map_by_id = dict()

        self.main_layout = layouts.GridLayout(margins=(9, 9, 9, 9))
        self.setLayout(self.main_layout)

    @staticmethod
    def check_section_container_lookup_map_by_id(dock_container, container_id):
        return container_id in dock_container._section_container_lookup_map_by_id

    @staticmethod
    def check_section_container_lookup_map_by_name(dock_container, name):
        return name in dock_container._section_container_lookup_map_by_name

    @staticmethod
    def check_section_widget_lookup_map_by_id(dock_container, widget_id):
        return widget_id in dock_container._section_widget_lookup_map_by_id

    def floatings(self):
        return self._floatings

    def is_empty(self):
        """
        Returns whether the container has sections stored in it or not
        :return: bool
        """

        return len(self._sections) <= 0

    def register_section_content(self, section_content):
        """
        Registers the given section content
        :param section_content: DockSectionContent
        """

        # TODO: We should check if the register is correct (no duplication unique names or ids)
        # TODO: And abort the process is something is wrong?

        self._section_container_lookup_map_by_id[section_content.uid()] = section_content
        self._section_container_lookup_map_by_name[section_content.unique_name()] = section_content

    def register_section_widget(self, section_widget):
        """
        Registers the given section widget
        :param section_widget: DockSectionWidget
        """

        # TODO: We should check if the register is correct (no duplication unique names or ids)
        # TODO: And abort the process is something is wrong?

        self._section_widget_lookup_map_by_id[section_widget.uid()] = section_widget
        self._sections.append(section_widget)

    def clean(self):
        self._sections.clear()
        del self._sections[:]

        self._floatings.clear()
        del self._floatings[:]

        self._sc_lookup_map_by_id.clear()
        self._sc_lookup_map_by_name.clear()
        self._sc_lookup_map_by_id.clear()

    def new_section_widget(self):
        """
        Creates a new DockSectionContentWidget from scratch
        :return: DockSectionContentWidget
        """

        section_widget = DockSectionWidget(container_widget=self)
        self.register_section_widget(section_widget)

        return section_widget

    def section_at(self, pos):
        """
        Returns the section, if exists, that is located on top of the cursor
        :param pos: QPoint
        :return: DockSectionWidget
        """

        g_pos = self.mapToGlobal(pos)
        for i in range(len(self._sections)):
            section_widget = self._sections[i]
            if section_widget.rect().contains(section_widget.mapFromGlobal(g_pos)):
                return section_widget

        return None

    def drop_content_outer_helper(self, parent_layout, internal_data, orientation, append):
        if parent_layout is None:
            return

        section_widget = self.new_section_widget()
        section_widget.add_data_content(internal_data=internal_data, auto_activate=True)

        old_splitter = find_inmediate_splitter(widget=self)
        if not old_splitter:
            sp = new_dock_splitter(orientation=orientation)
            if self.main_layout.count() > 0:
                LOGGER.warning('Still items in layout. This should never happen!')
                layout_item = self.main_layout.takeAt(0)
                layout_item.deleteLater()
                del layout_item

            self.main_layout.addWidget(sp)
            sp.addWidget(section_widget)
        elif old_splitter.orientation() == orientation or old_splitter.count() == 1:
            old_splitter.setOrientation(orientation)
            if append:
                old_splitter.addWidget(section_widget)
            else:
                old_splitter.insertWidget(0, section_widget)
        else:
            sp = new_dock_splitter(orientation=orientation)
            if append:
                try:
                    layout_item = self.main_layout.replaceWidget(old_splitter, sp)
                    sp.addWidget(old_splitter)
                    sp.addWidget(section_widget)
                    layout_item.deleteLater()
                    del layout_item
                except Exception:
                    index = self.main_layout.indexOf(old_splitter)
                    layout_item = self.main_layout.takeAt(index)
                    self.main_layout.addWidget(old_splitter)
                    self.main_layout.addWidget(section_widget)
                    layout_item.deleteLater()
                    del layout_item
            else:
                try:
                    sp.addWidget(section_widget)
                    item_layout = self.main_layout.replaceWidget(old_splitter, sp)
                    sp.addWidget(old_splitter)
                    item_layout.deleteLater()
                    del item_layout
                except Exception:
                    sp.addWidget(section_widget)
                    index = self.main_layout.indexOf(old_splitter)
                    item_layout = self.main_layout.takeAt(index)
                    self.main_layout.addWidget(sp)
                    sp.addWidget(old_splitter)
                    item_layout.deleteLater()
                    del item_layout

        return section_widget

    def drop_content(self, internal_data, target_section_widget, drop_area, auto_active):
        """
        Adds and fills DockSectionContent based on the given data and positioned in the given drop.DropArea
        :param internal_data: DockInternalContentData, stores all dock information to add to the DockContainer
        :param target_section_widget: DockSectionWidget
        :param drop_area: drop.DropArea, area where we want to drop the new DockSectionContent
        :param auto_active: bool, whether if the new DockSectionContent should be visible or not by default
        :return: DropSectionContent, new DropSectionContent added
        """

        ret = None

        # If the DockContainer has no DockSectionContents added to it, we create a new one from scratch and is loaded
        # automatically in the center of the DockContainer
        if self.is_empty():
            LOGGER.debug('DockContainer is empty! Creating DockSectionWidget from scratch ...')
            target_section_widget = self.new_section_widget()
            self.add_section(target_section_widget)
            drop_area = drop.DropArea.CenterDropArea

        # If no the container is not empty and we do not specify a DockSectionWidget, we create a new one
        # from scratch dropped in the given drop area
        if not target_section_widget:
            if drop_area == drop.DropArea.TopDropArea:
                ret = self.drop_content_outer_helper(
                    parent_layout=self.main_layout,
                    internal_data=internal_data, orientation=Qt.Vertical, append=False)
            elif drop_area == drop.DropArea.RightDropArea:
                ret = self.drop_content_outer_helper(
                    parent_layout=self.main_layout,
                    internal_data=internal_data, orientation=Qt.Horizontal, append=True)
            elif drop_area == drop.DropArea.CenterDropArea or drop_area == drop.DropArea.BottomDropArea:
                ret = self.drop_content_outer_helper(
                    parent_layout=self.main_layout,
                    internal_data=internal_data, orientation=Qt.Vertical, append=True)
            elif drop_area == drop.DropArea.LeftDropArea:
                self.drop_content_outer_helper(
                    parent_layout=self.main_layout,
                    internal_data=internal_data, orientation=Qt.Horizontal, append=False)
            else:
                return None
            return ret

        # We loop through all the widgets of the DockSectionWidget and we get its parent (splitter)
        target_section_splitter = find_parent_splitter(target_section_widget)
        if target_section_splitter is None:
            LOGGER.warning('DockSectionWidget is not parented to any widget. This cannot happen!')
            return

        # Add the DockSectionWidget into its corresponding area
        if drop_area == drop.DropArea.TopDropArea:
            section_widget = self.new_section_widget()
            section_widget.add_data_content(internal_data, True)
            if target_section_splitter.orientation() == Qt.Vertical:
                index = target_section_splitter.indexOf(target_section_widget)
                target_section_splitter.insertWidget(index, section_widget)
            else:
                index = target_section_splitter.indexOf(target_section_widget)
                s = new_dock_splitter(Qt.Vertical)
                s.addWidget(section_widget)
                s.addWidget(target_section_widget)
                target_section_splitter.insertWidget(index, s)
            ret = section_widget
        elif drop_area == drop.DropArea.RightDropArea:
            section_widget = self.new_section_widget()
            section_widget.add_data_content(internal_data, True)
            if target_section_splitter.orientation() == Qt.Horizontal:
                index = target_section_splitter.indexOf(target_section_widget)
                target_section_splitter.insertWidget(index + 1, section_widget)
            else:
                index = target_section_splitter.indexOf(target_section_widget)
                s = new_dock_splitter(Qt.Horizontal)
                s.addWidget(target_section_widget)
                s.addWidget(section_widget)
                target_section_splitter.insertWidget(index, s)
            ret = section_widget
        elif drop_area == drop.DropArea.BottomDropArea:
            section_widget = self.new_section_widget()
            section_widget.add_data_content(internal_data, True)
            if target_section_splitter.orientation() == Qt.Vertical:
                index = target_section_splitter.indexOf(target_section_widget)
                target_section_splitter.insertWidget(index + 1, section_widget)
            else:
                index = target_section_splitter.indexOf(target_section_widget)
                s = new_dock_splitter(Qt.Vertical)
                s.addWidget(target_section_widget)
                s.addWidget(section_widget)
                target_section_splitter.insertWidget(index, s)
            ret = section_widget
        elif drop_area == drop.DropArea.LeftDropArea:
            section_widget = self.new_section_widget()
            section_widget.add_data_content(internal_data, True)
            if target_section_splitter.orientation() == Qt.Horizontal:
                index = target_section_splitter.indexOf(target_section_widget)
                target_section_splitter.insertWidget(index, section_widget)
            else:
                s = new_dock_splitter(Qt.Horizontal)
                s.addWidget(section_widget)
                index = target_section_splitter.indexOf(target_section_widget)
                target_section_splitter.insertWidget(index, s)
                s.addWidget(target_section_widget)
            ret = section_widget
        elif drop_area == drop.DropArea.CenterDropArea:
            target_section_widget.add_data_content(internal_data, auto_activate=auto_active)
            ret = target_section_widget

        return ret

    def add_section(self, section_widget):
        """
        Adds the given DockSectionContentWidget to the Main Splitter DockContainer
        If the splitter does not exists, it's created automatically
        :param section_widget: DockSectionContentWidget
        """

        if not section_widget:
            LOGGER.warning('Impossible to add no valid DockSectionContentWidget!')
            return

        # If DockContainer's Main Splitter does not exists, we create it ...
        if not self._splitter:
            self._splitter = new_dock_splitter(orientation=self._orientation)
            self.main_layout.addWidget(self._splitter, 0, 0)

        # Check if the given DockSectionContentWidget has been already to the splitter.
        # If that's the case we do not add it
        if self._splitter.indexOf(section_widget) != -1:
            LOGGER.warning('Section has already been added!')
            return

        # If everything is ok, we add the DockSectionContentWidget to the splitter
        self._splitter.addWidget(section_widget)

    def take_content(self, section_content, internal_data):
        if section_content is None:
            return

        found = False

        # Search in sections
        for i in range(len(self._sections)):
            found = self._sections[i].take_content(section_content.uid(), internal_data)

        # Search in floating widgets
        if not found:
            for i in range(len(self._floatings)):
                found = self._floatings[i].content().uid() == section_content.uid()
                if found:
                    self._floatings[i].take_content(internal_data)

        # Search in hidden items
        if not found:
            if section_content.uid() in self._hidden_section_contents:
                hidden_section_item = self._hidden_section_contents[section_content.uid()]
                internal_data = hidden_section_item.data
                found = True

        return found

    def outer_top_drop_rect(self):
        r = self.rect()
        h = r.height() / 100 * 5
        return QRect(r.left(), r.top(), r.width(), h)

    def outer_right_drop_rect(self):
        r = self.rect()
        w = r.width() / 100 * 5
        return QRect(r.right() - w, r.top(), w, r.height())

    def outer_bottom_drop_rect(self):
        r = self.rect()
        w = r.width() / 100 * 5
        return QRect(r.right() - w, r.top(), w, r.height())

    def outer_left_drop_rect(self):
        r = self.rect()
        w = r.width() / 100 * 5
        return QRect(r.left(), r.top(), w, r.height())

    def contents(self):
        section_containers = self._section_container_lookup_map_by_id.values()
        contents = list()
        for i in range(len(section_containers)):
            section_content = section_containers[i]
            if section_content:
                contents.append(section_content)

        return contents

    def add_section_content(self, section_content, section_widget, drop_area):
        """
        Adds the given section content to this DockContainer
        If DockSectionWidget is not specified, a new one is created for the given DockSectionContent
        :param section_content: DockSectionContent we want to add to this DockContainer
        :param section_widget: DockSectionWidget that will be managed by the DockSectionContent
        :param drop_area: drop.DropArea, area where we want to add the new DockSectionContent into
        :return:
        """

        # If no DockSectionWidget is given, we abort the process
        if section_content is None:
            LOGGER.warning('Cannot add new DockSectionWidget if no one is specified')
            return

        # Create auxiliary data class to store dock data information
        # 1. DockSectionContent : stored from the given one
        # 2. Title Widget : created from scratch and parented to the DockSectionContent
        # 3. DockSectionContentWidget : create from scratch and parent to the DockSectionContent
        data = DockInternalContentData()
        data.content = section_content
        data.title_widget = DockSectionTitleWidget(section_content)
        data.content_widget = DockSectionContentWidget(section_content)

        # When the DOckSectionTitLeWidget is pressed  we notify that the tab have changed
        data.title_widget.activeTabChanged.connect(self._on_active_tab_changed)

        return self.drop_content(
            internal_data=data, target_section_widget=section_widget, drop_area=drop_area, auto_active=False)

    def show_section_content(self, section_content):
        if section_content is None:
            return

        # Search Section Content in floatings
        for i in range(len(self._floatings)):
            floating_widget = self._floatings[i]
            found = floating_widget.content().uid() == section_content.uid()
            if not found:
                continue
            floating_widget.setVisible(True)
            floating_widget._title_widget.setVisible(True)
            floating_widget._content_widget.setVisible(True)
            self.sectionContentVisibilityChanged.emit(section_content, True)
            return True

        # Search Section Content in hidden sections
        if section_content.uid() in self._hidden_section_contents:
            hidden_section_item = self._hidden_section_contents.pop(section_content.uid())
            hidden_section_item.data.title_widget.setVisible(True)
            hidden_section_item.data.contentWidget.setVisible(True)
            section_widget = None
            if hidden_section_item.preferred_section_id > 0:
                sw = self._section_container_lookup_map_by_id[hidden_section_item.preferred_section_id]
                if sw is not None:
                    sw.add_data_content(hidden_section_item.data, True)
                    self.sectionContentVisibilityChanged.emit(section_content, True)
                    return True
            elif len(self._sections) and self._sections[0] is not None:
                section_widget.add_data_content(hidden_section_item.data, True)
                self.sectionContentVisibilityChanged.emit(section_content, True)
                return True
            else:
                sw = self.new_section_widget()
                self.add_section(section_widget=sw)
                sw.add_data_content(hidden_section_item.data, True)
                self.sectionContentVisibilityChanged.emit(section_content, True)
                return True

        # Already visible?
        LOGGER.warning('Unable to show SectionContent, do not know where (already visible)?')
        return False

    def hide_section_content(self, section_content):
        """
        Hides the section content
        :param section_content: DockSectionContent you want to hide
        :return:
        """

        if section_content is None:
            return

        # Search Section Content in floatings
        for i in range(len(self._floatings)):
            found = self._floatings[i].content().uid() == section_content.uid()
            if not found:
                continue
            self._floatings[i].setVisible(False)
            self.sectionContentVisibilityChanged.emit(section_content, False)
            return True

        # Search Section Content in sections
        for i in range(len(self._sections)):
            section_widget = self._sections[i]
            found = section_widget.index_of_content(section_content) >= 0
            if not found:
                continue

            hidden_selection_item = DockHiddenSectionItem()
            hidden_selection_item.preferred_section_id = section_widget.uid()
            hidden_selection_item.preferred_section_index = section_widget.index_of_content(section_content)
            hidden_selection_item.data = section_widget.take_content(section_content.uid())
            if not hidden_selection_item.data:
                return False

            hidden_selection_item.data.title_widget.setVisible(False)
            hidden_selection_item.data.content_widget.setVisible(False)
            self._hidden_section_contents[section_content.uid()] = hidden_selection_item

            if section_widget.is_empty():
                section_widget.clean()
                section_widget.deleteLater()
                del section_widget
                delete_empty_splitter(self)

            self.sectionContentVisibilityChanged.emit(section_content, False)
            return True

        # Search Section Content in hidden elements
        if section_content.uid() in self._hidden_section_contents:
            return True

        LOGGER.error('Unable to hide SectionContent, do not know this one!')
        return False

    def raise_section_content(self, section_content):
        if section_content is None:
            return

        # Search Section Content in sections
        for i in range(len(self._sections)):
            sw = self._sections[i]
            index = sw.index_of_content(section_content)
            if index < 0:
                continue
            sw.set_current_index(index)
            return True

        # Search Section Content in floatins
        for i in range(len(self._floatings)):
            fw = self._floatings[i]
            if fw.content().uid() != section_content.uid():
                continue
            fw.setVisible(True)
            fw.raise_()
            return True

        # Search Section Content in hidden
        if section_content.uid() in self._hidden_section_contents:
            return self.show_section_content(section_content)

        LOGGER.error('Unable hide SectionContent, do not know this one!')
        return False

    def remove_section_content(self, section_content):
        if section_content is None:
            return

        if not self.hide_section_content(section_content):
            return False

        found = False
        for i in range(len(self._floatings)):
            fw = self._floatings[i]
            data = DockInternalContentData()
            found = fw.take_content(data)
            if not found:
                continue

            self._floatings.remove(fw)
            fw.clean()
            fw.deleteLater()
            del fw
            data.title_widget.clean()
            data.title_widget.deleteLater()
            del data.title_widget
            data.content_widget.clean()
            data.content_widget.deleteLater()
            del data.content_widget

        if section_content.uid() not in self._hidden_section_contents:
            LOGGER.error('Something went wrong ... The content should have been there!')
            return False

        hidden_selection_item = self._hidden_section_contents.pop(section_content.uid())
        hidden_selection_item.data.title_widget.deleteLater()
        del hidden_selection_item.data.title_widget
        hidden_selection_item.data.content_widget.deleteLater()
        del hidden_selection_item.data.content_widget

        try:
            section_content.clean()
            del section_content
        except Exception:
            LOGGER.warning('Maybe something went wrong when deleting section ...')

        if self.is_empty():
            self._splitter.deleteLater()
            del self._splitter
            self._splitter = None

        return True

    def is_section_content_visible(self, section_content):
        if section_content is None:
            return

        # Search Section Content in floatings
        for i in range(len(self._floatings)):
            found = self._floatings[i].content().uid() == section_content.uid()
            if not found:
                continue
            return self._floatings[i].isVisible()

        # Search Section Content in sections
        for i in range(len(self._sections)):
            sw = self._sections[i]
            index = sw.index_of_content(section_content)
            if index < 0:
                continue
            return True

        # Search Section Content in hidden
        if section_content.uid() in self._hidden_section_contents:
            return False

        LOGGER.warning(
            'SectionContent is not a part of this ContainerWidget: {}'.format(section_content.unique_name()))
        return False

    def drop_overlay(self):
        return self._drop_overlay

    def _on_active_tab_changed(self):
        section_title_widget = self.sender()
        if section_title_widget:
            self.activeTabChanged.emit(section_title_widget.content(), section_title_widget.is_active_tab())

    def _on_action_toggle_section_content_visibility(self, visible):
        a = self.sender()
        if not a:
            return
        uid = a.property('uid').toInt()
        section_content = self._section_container_lookup_map_by_id[uid]
        if not section_content:
            LOGGER.error('Cannot bind content by ID'.format(uid))
            return
        if visible:
            self.show_section_content(section_content)
        else:
            self.hide_section_content(section_content)



import random

import tpDcc
from tpDcc.libs.qt.core import mixin, theme, menu
from tpDcc.libs.qt.widgets import layouts, combobox, dividers, label


class ComboBoxExample(tpDcc.Window, mixin.FieldMixin):
    def __init__(self):
        super(ComboBoxExample, self).__init__(title='Example for ComboBoxes')

    def ui(self):
        super(ComboBoxExample, self).ui()

        def _get_letters():
            data = ['A', 'B', 'C', 'D', 'E', 'F']
            start = random.randint(0, len(data))
            end = random.randint(start, len(data))
            return data[start:end]

        a = [{'children': [{'value': '\u6545\u5bab', 'label': '\u6545\u5bab'},
                           {'value': '\u5929\u575b', 'label': '\u5929\u575b'},
                           {'value': '\u738b\u5e9c\u4e95', 'label': '\u738b\u5e9c\u4e95'}],
              'value': u'\u5317\u4eac',
              'label': u'\u5317\u4eac'},
             {'children': [{'children': [{'value': '\u592b\u5b50\u5e99', 'label': '\u592b\u5b50\u5e99'}],
                            'value': '\u5357\u4eac',
                            'label': '\u5357\u4eac'},
                           {'children': [{'value': '\u62d9\u653f\u56ed', 'label': '\u62d9\u653f\u56ed'},
                                         {'value': '\u72ee\u5b50\u6797', 'label': '\u72ee\u5b50\u6797'}],
                            'value': '\u82cf\u5dde',
                            'label': '\u82cf\u5dde'}],
              'value': '\u6c5f\u82cf',
              'label': '\u6c5f\u82cf'}]

        current_theme = self.theme()

        self.register_field('button1_selected', 'A')
        self.register_field('button2_selected', 'A')
        self.register_field('button3_selected', '')
        self.register_field('button4_selected', '')
        self.register_field('button5_selected', '')

        menu1 = menu.BaseMenu(parent=self)
        menu1.set_data(['A', 'B', 'C', 'D'])
        menu2 = menu.BaseMenu(exclusive=False, parent=self)
        menu2.set_data(['A', 'B', 'C', 'D'])
        menu3 = menu.BaseMenu(parent=self)
        menu3.set_load_callback(_get_letters)
        menu4 = menu.BaseMenu(parent=self, cascader=True)
        menu4.set_data(a)
        menu5 = menu.BaseMenu(exclusive=False, parent=self)
        menu5.set_data(['A', 'B', 'C', 'D'])

        size_list = [
            ('Small', current_theme.small if current_theme else theme.Theme.Sizes.SMALL),
            ('Medium', current_theme.small if current_theme else theme.Theme.Sizes.MEDIUM),
            ('Large', current_theme.small if current_theme else theme.Theme.Sizes.LARGE)
        ]

        size_lyt = layouts.HorizontalLayout()
        for lbl, size in size_list:
            combo_box = combobox.BaseComboBox()
            combo_box.theme_size = size
            combo_box.set_menu(menu1)
            size_lyt.addWidget(combo_box)
            self.bind('button1_selected', combo_box, 'value', signal='valueChanged')

        select2 = combobox.BaseComboBox()
        select2.set_menu(menu2)
        self.bind('button2_selected', select2, 'value', signal='valueChanged')

        select3 = combobox.BaseComboBox()
        select3.set_menu(menu3)
        self.bind('button3_selected', select3, 'value', signal='valueChanged')

        select4 = combobox.BaseComboBox()
        select4.set_menu(menu4)
        select4.set_formatter(lambda x: ' / '.join(x))
        self.bind('button4_selected', select4, 'value', signal='valueChanged')

        select5 = combobox.BaseComboBox()
        select5.set_menu(menu5)
        select5.set_formatter(lambda x: ' & '.join(x))
        self.bind('button5_selected', select5, 'value', signal='valueChanged')

        sub_lyt1 = layouts.HorizontalLayout()
        sub_lyt1.addWidget(label.BaseLabel('Normal selection with different sizes'))
        sub_lyt1.addLayout(size_lyt)
        sub_lyt1.addStretch()
        sub_lyt2 = layouts.HorizontalLayout()
        sub_lyt2.addWidget(label.BaseLabel('Mutliple options'))
        sub_lyt2.addWidget(select2)
        sub_lyt2.addStretch()
        sub_lyt3 = layouts.HorizontalLayout()
        sub_lyt3.addWidget(label.BaseLabel('Dynamic options'))
        sub_lyt3.addWidget(select3)
        sub_lyt3.addStretch()
        sub_lyt4 = layouts.HorizontalLayout()
        sub_lyt4.addWidget(label.BaseLabel('Cascader Selection'))
        sub_lyt4.addWidget(select4)
        sub_lyt4.addStretch()
        sub_lyt5 = layouts.HorizontalLayout()
        sub_lyt5.addWidget(label.BaseLabel('Custom Screen'))
        sub_lyt5.addWidget(select5)
        sub_lyt5.addStretch()

        self.main_layout.addWidget(dividers.Divider('Select'))
        self.main_layout.addLayout(sub_lyt1)
        self.main_layout.addLayout(sub_lyt2)
        self.main_layout.addLayout(sub_lyt3)
        self.main_layout.addWidget(dividers.Divider('Custom Format'))
        self.main_layout.addLayout(sub_lyt4)
        self.main_layout.addLayout(sub_lyt5)
        self.main_layout.addStretch()


def run():
    return ComboBoxExample()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains classes to create different kind of line edits
"""

from __future__ import print_function, division, absolute_import

from functools import partial

from Qt.QtCore import Qt, Signal, Property, QTimer
from Qt.QtWidgets import QApplication, QLineEdit, QTextEdit
from Qt.QtGui import QDoubleValidator, QIntValidator

from tpDcc.libs.resources.core import theme
from tpDcc.libs.qt.core import contexts as qt_contexts
from tpDcc.libs.qt.widgets import layouts, buttons, browser


@theme.mixin
class BaseLineEdit(QLineEdit, object):
    """
     Basic line edit
     """

    delayTextChanged = Signal(str)

    def __init__(self, text='', input_mode=None, parent=None):
        super(BaseLineEdit, self).__init__(text, parent)

        self._prefix_widget = None
        self._suffix_widget = None
        self._size = self.theme_default_size()

        self._main_layout = layouts.HorizontalLayout()
        self._main_layout.setContentsMargins(0, 0, 0, 0)
        self._main_layout.addStretch()
        self.setLayout(self._main_layout)

        self.setProperty('history', self.property('text'))
        self.setTextMargins(2, 0, 2, 0)

        if input_mode == 'float':
            self.setValidator(QDoubleValidator())
        elif input_mode == 'int':
            self.setValidator(QIntValidator())

        self._delay_timer = QTimer()
        self._delay_timer.setInterval(500)
        self._delay_timer.setSingleShot(True)
        self._delay_timer.timeout.connect(self._on_delay_text_changed)

    # =================================================================================================================
    # PROPERTIES
    # =================================================================================================================

    def _get_text(self):
        return self.text()

    def _set_text(self, value):
        with qt_contexts.block_signals(self):
            self.setText(value)

    def _get_size(self):
        """
        Returns the spin box height size
        :return: float
        """

        return self._size

    def _set_size(self, value):
        """
        Sets spin box height size
        :param value: float
        """

        self._size = value
        if hasattr(self._prefix_widget, 'theme_size'):
            self._prefix_widget.theme_size = self._size
        if hasattr(self._suffix_widget, 'theme_size'):
            self._suffix_widget.theme_size = self._size
        self.style().polish(self)

    theme_size = Property(int, _get_size, _set_size)
    line_text = Property(str, _get_text, _set_text)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def setText(self, text):
        """
        Overrides base QLineEdit setText base function.
        Save history
        :param text: str
        """

        self.setProperty('history', '{}\n{}'.format(self.property('history'), text))
        return super(BaseLineEdit, self).setText(text)

    def clear(self):
        """
        Overrides base QLineEdit clear function
        :return:
        """

        self.setProperty('history', '')
        return super(BaseLineEdit, self).clear()

    def keyPressEvent(self, event):
        """
        Overrides base QLineEdit keyPressEvent function
        :param event: QKeyEvent
        """

        if event.key() not in [Qt.Key_Enter, Qt.Key_Tab]:
            if self._delay_timer.isActive():
                self._delay_timer.stop()
            self._delay_timer.start()
        super(BaseLineEdit, self).keyPressEvent(event)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def set_delay_duration(self, ms):
        """
        Sets the delay timer duration
        :param ms: float
        """

        self._delay_timer.setInterval(ms)

    def get_prefix_widget(self):
        """
        Returns prefix widget for user to edit
        :return: QWidget
        """

        return self._prefix_widget

    def set_prefix_widget(self, widget):
        """
        Sets the edit line left start widget
        :param widget: QWidget
        :return: QWidget
        """

        if self._prefix_widget:
            index = self._main_layout.indexOf(self._prefix_widget)
            self._main_layout.takeAt(index)
            self._prefix_widget.setVisible(False)
            self._prefix_widget.deleteLater()

        widget.setProperty('combine', 'horizontal')
        widget.setProperty('position', 'left')
        if hasattr(widget, 'theme_size'):
            widget.them_size = self.theme_size

        margin = self.textMargins()
        margin.setLeft(margin.left() + widget.width())
        self.setTextMargins(margin)

        self._main_layout.insertWidget(0, widget)
        self._prefix_widget = widget

        return widget

    def get_suffix_widget(self):
        """
        Returns suffix widget for user to edit
        :return: QWidget
        """

        return self._suffix_widget

    def set_suffix_widget(self, widget):
        """
        Sets the edit line right start widget
        :param widget: QWidget
        :return: QWidget
        """

        if self._suffix_widget:
            index = self._main_layout.indexOf(self._suffix_widget)
            self._main_layout.takeAt(index)
            self._suffix_widget.setVisible(False)
            self._suffix_widget.deleteLater()

        widget.setProperty('combine', 'horizontal')
        widget.setProperty('position', 'right')
        if hasattr(widget, 'theme_size'):
            widget.them_size = self.theme_size

        margin = self.textMargins()
        margin.setRight(margin.right() + widget.width())
        self.setTextMargins(margin)

        self._main_layout.addWidget(widget)
        self._prefix_widget = widget

        return widget

    def search(self):
        """
        Adds a search icon button for line edit
        :return: self
        """

        prefix_btn = buttons.BaseToolButton().image('search').icon_only()
        suffix_btn = buttons.BaseToolButton().image('close').icon_only()
        suffix_btn.clicked.connect(self.clear)
        self.set_prefix_widget(prefix_btn)
        self.set_suffix_widget(suffix_btn)
        self.setPlaceholderText('Enter keyword to search ...')

        return self

    def search_engine(self, text='Search'):
        """
        Adds a search push button to line edit
        :param text: str
        :return: self
        """

        _suffix_btn = buttons.BaseButton(text).primary()
        _suffix_btn.clicked.connect(self.returnPressed)
        _suffix_btn.setFixedWidth(100)
        self.set_suffix_widget(_suffix_btn)
        self.setPlaceholderText('Enter keyword to search ...')

        return self

    def file(self, filters=None):
        """
        Adds a ClickBrowserFileToolButton to line edit
        :param filters:
        :return: self
        """

        _suffix_btn = browser.ClickBrowserFileToolButton()
        _suffix_btn.fileChanged.connect(self.setText)
        _suffix_btn.filters = filters
        self.textChanged.connect(_suffix_btn.set_path)
        self.set_suffix_widget(_suffix_btn)
        self.setPlaceholderText('Click button to browse files')

        return self

    def save_file(self, filters=None):
        """
        Adds a ClickSaveFileToolButton to line edit
        :param filters:
        :return: self
        """

        _suffix_button = browser.ClickSaveFileToolButton()
        _suffix_button.fileChanged.connect(self.setText)
        _suffix_button.filters = filters or list()
        self.textChanged.connect(_suffix_button.set_path)
        self.set_suffix_widget(_suffix_button)
        self.setPlaceholderText('Click button to set save file')

        return self

    def folder(self):
        """
        Adds a ClickBrowserFileToolButton to line edit
        :return: self
        """

        _suffix_btn = browser.ClickBrowserFolderToolButton()
        _suffix_btn.folderChanged.connect(self.setText)
        self.textChanged.connect(_suffix_btn.set_path)
        self.set_suffix_widget(_suffix_btn)
        self.setPlaceholderText('Click button to browse folder')

        return self

    def error(self):
        """
        Shows error in line edit with red style
        :return: self
        """

        def _on_show_detail(self):
            dlg = QTextEdit(self)
            dlg.setReadOnly(True)
            geo = QApplication.desktop().screenGeometry()
            dlg.setGeometry(geo.width() / 2, geo.height() / 2, geo.width() / 4, geo.height() / 4)
            dlg.setWindowTitle('Error Detail Information')
            dlg.setText(self.property('history'))
            dlg.setWindowFlags(Qt.Dialog)
            dlg.show()

        self.setProperty('theme_type', 'error')
        self.setReadOnly(True)
        _suffix_btn = buttons.BaseToolButton().image('delete_message').icon_only()
        _suffix_btn.clicked.connect(partial(_on_show_detail, self))
        self.set_suffix_widget(_suffix_btn)
        self.setPlaceholderText('Error information will be here ...')

        return self

    def tiny(self):
        """
        Sets line edit to tiny size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.tiny if widget_theme else theme.Theme.Sizes.TINY

        return self

    def small(self):
        """
        Sets line edit to small size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.small if widget_theme else theme.Theme.Sizes.SMALL

        return self

    def medium(self):
        """
        Sets line edit to medium size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.medium if widget_theme else theme.Theme.Sizes.MEDIUM

        return self

    def large(self):
        """
        Sets line edit to large size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.large if widget_theme else theme.Theme.Sizes.LARGE

        return self

    def huge(self):
        """
        Sets line edit to huge size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.huge if widget_theme else theme.Theme.Sizes.HUGE

        return self

    def password(self):
        """
        Sets line edit password mode
        """

        self.setEchoMode(QLineEdit.Password)

        return self

    # =================================================================================================================
    # CALLBACKS
    # =================================================================================================================

    def _on_delay_text_changed(self):
        """
        Internal callback function that is called when delay timer is completed
        """

        self.delayTextChanged.emit(self.text())


class StyledLineEdit(QLineEdit, object):
    """
    Styled line edit that takes a different color if it's empty
    """

    def __init__(self, default='', off_color=(125, 125, 125), on_color=(255, 255, 255), parent=None):
        super(StyledLineEdit, self).__init__(parent=parent)

        self._value = ''
        self._default = ''
        self._off_color = off_color
        self._on_color = on_color

        self.set_default(default)
        self.textChanged.connect(self._on_change)

    def get_value(self):
        if self.text() == self._default:
            return ''
        else:
            return self._value

    def set_value(self, value):
        self._value = value

    value = property(get_value, set_value)

    def focusInEvent(self, event):
        if self.text() == self._default:
            self.setText('')
            self.setStyleSheet(self._get_on_style())

    def focusOutEvent(self, event):
        if self.text() == '':
            self.setText(self._default)
            self.setStyleSheet(self._get_off_style())

    def set_default(self, text):
        self.setText(text)
        self._default = text
        self.setStyleSheet(self._get_off_style())

    def _get_on_style(self):
        return 'QLineEdit{color:rgb(%s, %s, %s);}' % (self._on_color[0], self._on_color[1], self._on_color[2])

    def _get_off_style(self):
        return 'QLineEdit{color:rgb(%s, %s, %s);}' % (self._off_color[0], self._off_color[1], self._off_color[2])

    def _on_change(self, text):
        if text != self._default:
            self.setStyleSheet(self._get_on_style())
            self._value = text
        else:
            self.setStyleSheet(self._get_off_style())


class ClickLineEdit(QLineEdit, object):
    """
    Custom QLineEdit that becomes editable on click or double click
    """

    def __init__(self, text, single=False, double=False, pass_through_click=True):
        super(ClickLineEdit, self).__init__(text)

        self.setReadOnly(True)
        self._editing_style = self.styleSheet()
        self._default_style = "QLineEdit {border: 0;}"
        self.setStyleSheet(self._default_style)
        self.setContextMenuPolicy(Qt.NoContextMenu)

        if single:
            self.mousePressEvent = self.editEvent
        else:
            if pass_through_click:
                self.mousePressEvent = self._mouse_click_pass_through
        if double:
            self.mouseDoubleClickEvent = self.editEvent
        else:
            if pass_through_click:
                self.mousePressEvent = self._mouse_click_pass_through

        self.editingFinished.connect(self._on_edit_finished)

    def focusOutEvent(self, event):
        super(ClickLineEdit, self).focusOutEvent(event)
        self._on_edit_finished()

    def mousePressEvent(self, event):
        event.ignore()

    def mouseReleaseEvent(self, event):
        event.ignore()

    def editEvent(self, event):
        self.setStyleSheet(self._editing_style)
        self.selectAll()
        self.setReadOnly(False)
        self.setFocus()
        event.accept()

    def _mouse_click_pass_through(self, event):
        event.ignore()

    def _on_edit_finished(self):
        self.setReadOnly(True)
        self.setStyleSheet(self._default_style)
        self.deselect()


class BaseAttrLineEdit(QLineEdit, object):
    attr_type = None
    valueChanged = Signal()

    def __init__(self, parent=None):
        super(BaseAttrLineEdit, self).__init__(parent=parent)

        self.returnPressed.connect(self.update)
        self.editingFinished.connect(self.update)

    def get_value(self):
        return None

    value = property(get_value)


class FloatLineEdit(BaseAttrLineEdit, object):
    attr_type = 'float'
    valueChanged = Signal(float)

    def __init__(self, parent=None):
        super(FloatLineEdit, self).__init__(parent=parent)

    # region Override Properties
    def get_value(self):
        if not self.text():
            return 0.0
        return float(self.text())

    value = property(get_value)

    def setText(self, text):
        super(FloatLineEdit, self).setText('%.2f' % float(text))

    def update(self):
        if self.text():
            self.setText(self.text())
        super(FloatLineEdit, self).update()
        self.valueChanged.emit(float(self.text()))


class IntLineEdit(QLineEdit, object):
    attr_type = 'int'
    valueChanged = Signal(int)

    def __init__(self, parent=None):
        super(IntLineEdit, self).__init__(parent=parent)

    def get_value(self):
        if not self.text():
            return 0
        return int(self.text())

    value = property(get_value)

    def setText(self, text):
        super(IntLineEdit, self).setText('%s' % int(text))

    def update(self):
        if self.text():
            self.setText(self.text())
        super(IntLineEdit, self).update()
        self.valueChanged.emit(int(self.text()))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains extended Qt dialog classes
"""

from __future__ import print_function, division, absolute_import

import os
import string
import getpass
from functools import partial

from Qt.QtCore import Qt, QObject, Signal, QSize, QRectF, QFileInfo, QDir
from Qt.QtWidgets import QApplication, QGroupBox, QDesktopWidget, QDialog, QStatusBar, QSizePolicy, QGraphicsView
from Qt.QtWidgets import QGraphicsScene, QFileIconProvider, QListWidgetItem, QMessageBox, QInputDialog
from Qt.QtWidgets import QWidget, QLabel, QFrame, QPushButton, QSlider, QLineEdit, QComboBox, QCheckBox, QRadioButton

from tpDcc import dcc
from tpDcc.managers import resources
from tpDcc.abstract import dialog as abstract_dialog
from tpDcc.libs.resources.core import theme
from tpDcc.libs.qt.core import qtutils, animation, dragger, resizers
from tpDcc.libs.qt.widgets import layouts, dividers


class BaseDialog(QDialog, abstract_dialog.AbstractDialog):
    """
    Class to create basic Maya docked windows
    """

    dialogResizedFinished = Signal()
    closed = Signal()
    themeUpdated = Signal(object)
    styleReloaded = Signal(object)

    def __init__(self, **kwargs):

        parent = kwargs.get('parent', None) or dcc.get_main_window()
        super(BaseDialog, self).__init__(parent=parent)

        self._setup_resizers()

        title = kwargs.get('title', '')
        name = kwargs.get('name', title or self.__class__.__name__)
        width = kwargs.get('width', 600)
        height = kwargs.get('height', 800)
        show_on_initialize = kwargs.get('show_on_initialize', False)
        self._theme = None
        self._dpi = kwargs.get('dpi', 1.0)
        self._fixed_size = kwargs.get('fixed_size', False)
        self._has_title = kwargs.pop('has_title', False)
        self._size = kwargs.pop('size', (200, 125))
        self._title_pixmap = kwargs.pop('title_pixmap', None)
        self._toolset = kwargs.get('toolset', None)

        self.setObjectName(str(name))
        self.setFocusPolicy(Qt.StrongFocus)

        frameless = kwargs.get('frameless', True)
        self.set_frameless(frameless)

        self.ui()
        self.setup_signals()

        self._status_bar = QStatusBar(self)
        self.main_layout.addWidget(self._status_bar)
        if self._fixed_size:
            self._status_bar.hide()
        self._status_bar.hide()

        self.setWindowTitle(title)

        auto_load = kwargs.get('auto_load', True)
        if auto_load:
            self.load_theme()

        if show_on_initialize:
            self.center()
            self.show()

        if self._toolset:
            self.main_layout.addWidget(self._toolset)

        self.resize(width, height)

    def default_settings(self):
        """
        Returns default settings values
        :return: dict
        """

        return {
            "theme": {
                "accentColor": "rgb(80, 80, 80, 255)",
                "backgroundColor": "rgb(45, 45, 45, 255)",
            }
        }

    def load_theme(self):
        def_settings = self.default_settings()
        theme_settings = def_settings.get('theme', dict())

        self.set_theme_settings(theme_settings)

    def set_width_height(self, width, height):
        """
        Sets the width and height of the dialog
        :param width: int
        :param height: int
        """

        x = self.geometry().x()
        y = self.geometry().y()
        self.setGeometry(x, y, width, height)

    def is_frameless(self):
        """
        Returns whether or not frameless functionality for this window is enable or not
        :return: bool
        """

        return self.window().windowFlags() & Qt.FramelessWindowHint == Qt.FramelessWindowHint

    def set_frameless(self, flag):

        window = self.window()

        if flag and not self.is_frameless():
            window.setAttribute(Qt.WA_TranslucentBackground)
            if qtutils.is_pyside2() or qtutils.is_pyqt5():
                window.setWindowFlags(window.windowFlags() | Qt.FramelessWindowHint | Qt.NoDropShadowWindowHint)
            else:
                window.setWindowFlags(Qt.Window | Qt.FramelessWindowHint)
            self.set_resizer_active(True)
        elif not flag and self.is_frameless():
            window.setAttribute(Qt.WA_TranslucentBackground)
            if qtutils.is_pyside2() or qtutils.is_pyqt5():
                window.setWindowFlags(window.windowFlags() | Qt.FramelessWindowHint | Qt.NoDropShadowWindowHint)
            else:
                self.setWindowFlags(Qt.Window | Qt.FramelessWindowHint)
            self.set_resizer_active(False)

        # window.show()
        #
        # if self._frameless:
        #     self.setAttribute(Qt.WA_TranslucentBackground)
        #     if qtutils.is_pyside2():
        #         self.setWindowFlags(self.windowFlags() | Qt.FramelessWindowHint)
        #     else:
        #         self.setWindowFlags(Qt.Window | Qt.FramelessWindowHint)

    def center(self, to_cursor=False):
        """
        Move the dialog to the center of the current window
        """

        frame_geo = self.frameGeometry()
        if to_cursor:
            pos = QApplication.desktop().cursor().pos()
            screen = QApplication.desktop().screenNumber(pos)
            center_point = QApplication.desktop().screenGeometry(screen).center()
        else:
            center_point = QDesktopWidget().availableGeometry().center()
        frame_geo.moveCenter(center_point)
        self.move(frame_geo.topLeft())

    def fade_close(self):
        animation.fade_window(start=1, end=0, duration=400, object=self, on_finished=self.close)

    def get_main_layout(self):
        main_layout = layouts.VerticalLayout(spacing=0, margins=(2, 2, 2, 2))
        return main_layout

    def ui(self):

        self._central_layout = layouts.VerticalLayout(margins=(0, 0, 0, 0), spacing=0)
        self.setLayout(self._central_layout)

        self._top_widget = QWidget()
        self._top_layout = layouts.VerticalLayout(margins=(0, 0, 0, 0), spacing=0)
        self._top_widget.setLayout(self._top_layout)

        for r in self._resizers:
            r.setParent(self)
            r.windowResizedFinished.connect(self.dialogResizedFinished)
        self.set_resize_directions()

        self._dragger = dragger.DialogDragger(parent=self)
        self._dragger.setVisible(self.is_frameless())
        self._top_layout.addWidget(self._dragger)

        self.main_widget = DialogContents()
        self.main_layout = self.get_main_layout()
        self.main_widget.setLayout(self.main_layout)

        self.logo_view = QGraphicsView()
        self.logo_view.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.logo_view.setMaximumHeight(100)
        self._logo_scene = QGraphicsScene()
        self._logo_scene.setSceneRect(QRectF(0, 0, 2000, 100))
        self.logo_view.setScene(self._logo_scene)
        self.logo_view.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.logo_view.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.logo_view.setFocusPolicy(Qt.NoFocus)

        if self._has_title and self._title_pixmap:
            self._logo_scene.addPixmap(self._title_pixmap)
            self._top_layout.addWidget(self.logo_view)

        title_background_pixmap = self._get_title_pixmap()
        if self._has_title and title_background_pixmap:
            self._logo_scene.addPixmap(title_background_pixmap)
            self._top_layout.addWidget(self.logo_view)
        else:
            self.logo_view.setVisible(False)

        grid_layout = layouts.GridLayout()
        grid_layout.setHorizontalSpacing(0)
        grid_layout.setVerticalSpacing(0)
        grid_layout.setContentsMargins(0, 0, 0, 0)
        grid_layout.addWidget(self._top_widget, 1, 1, 1, 1)
        grid_layout.addWidget(self.main_widget, 2, 1, 1, 1)
        grid_layout.addWidget(self._top_left_resizer, 0, 0, 1, 1)
        grid_layout.addWidget(self._top_resizer, 0, 1, 1, 1)
        grid_layout.addWidget(self._top_right_resizer, 0, 2, 1, 1)
        grid_layout.addWidget(self._left_resizer, 1, 0, 2, 1)
        grid_layout.addWidget(self._right_resizer, 1, 2, 2, 1)
        grid_layout.addWidget(self._bottom_left_resizer, 3, 0, 1, 1)
        grid_layout.addWidget(self._bottom_resizer, 3, 1, 1, 1)
        grid_layout.addWidget(self._bottom_right_resizer, 3, 2, 1, 1)
        grid_layout.setColumnStretch(1, 1)
        grid_layout.setRowStretch(2, 1)

        # Shadow effect for dialog
        # BUG: This causes some rendering problems when using other shadow effects in child widgets of the window
        # BUG: Also detected problems when updating wigets (tree views, web browsers, etc)
        # https://bugreports.qt.io/browse/QTBUG-35196
        # shadow_effect = QGraphicsDropShadowEffect(self)
        # shadow_effect.setBlurRadius(qtutils.dpi_scale(15))
        # shadow_effect.setColor(QColor(0, 0, 0, 150))
        # shadow_effect.setOffset(qtutils.dpi_scale(0))
        # self.setGraphicsEffect(shadow_effect)

        self._central_layout.addLayout(grid_layout)

        for r in self._resizers:
            r.windowResizedFinished.connect(self.dialogResizedFinished)

        if self._size:
            self.resize(self._size[0], self._size[1])

    def statusBar(self):
        """
        Returns status bar of the dialog
        :return: QStatusBar
        """

        return self._status_bar

    def dpi(self):
        """
        Return the current dpi for the window
        :return: float
        """

        return float(self._dpi)

    def set_dpi(self, dpi):
        """
        Sets current dpi for the window
        :param dpi: float
        """

        self._dpi = dpi

    def theme(self):
        """
        Returns the current theme
        :return: Theme
        """

        if not self._theme:
            self._theme = theme.Theme()

        return self._theme

    def set_theme(self, theme):
        """
        Sets current window theme
        :param theme: Theme
        """

        self._theme = theme
        self._theme.updated.connect(self.reload_stylesheet)
        self.reload_stylesheet()

    def set_theme_settings(self, settings):
        """
        Sets the theme settings from the given settings
        :param settings: dict
        """

        # TODO: We should be able to give a dialog a theme to load
        # TODO: This will allow to setup specific themes if dialogs are launch from windows with
        # TODO: specific themes

        new_theme = resources.theme('default')
        if not new_theme:
            new_theme = theme.Theme()
        new_theme.set_settings(settings)
        self.set_theme(new_theme)

    def reload_stylesheet(self):
        """
        Reloads the stylesheet to the current theme
        """

        current_theme = self.theme()
        if not current_theme:
            return
        current_theme.set_dpi(self.dpi())
        stylesheet = current_theme.stylesheet()
        self.setStyleSheet(stylesheet)
        self.styleReloaded.emit(current_theme)

    def setup_signals(self):
        pass

    def set_logo(self, logo, offset=(930, 0)):
        logo = self._logo_scene.addPixmap(logo)
        logo.setOffset(offset[0], offset[1])

    def resizeEvent(self, event):
        # TODO: Take the width from the QGraphicsView not hardcoded :)
        if hasattr(self, 'logo_view'):
            self.logo_view.centerOn(1000, 0)
        return super(BaseDialog, self).resizeEvent(event)

    def closeEvent(self, event):
        self.closed.emit()
        event.accept()

    def setWindowIcon(self, icon):
        if self.is_frameless() or (hasattr(self, '_dragger') and self._dragger):
            self._dragger.set_icon(icon)
        super(BaseDialog, self).setWindowIcon(icon)

    def setWindowTitle(self, title):
        if self.is_frameless() or (hasattr(self, '_dragger') and self._dragger):
            self._dragger.set_title(title)
        super(BaseDialog, self).setWindowTitle(title)

    # ============================================================================================================
    # RESIZERS
    # ============================================================================================================

    def set_resizer_active(self, flag):
        """
        Sets whether resizers are enable or not
        :param flag: bool
        """

        if flag:
            for r in self._resizers:
                r.show()
        else:
            for r in self._resizers:
                r.hide()

    def set_resize_directions(self):
        """
        Sets the resize directions for the resizer widget of this window
        """

        self._top_resizer.set_resize_direction(resizers.ResizeDirection.Top)
        self._bottom_resizer.set_resize_direction(resizers.ResizeDirection.Bottom)
        self._right_resizer.set_resize_direction(resizers.ResizeDirection.Right)
        self._left_resizer.set_resize_direction(resizers.ResizeDirection.Left)
        self._top_left_resizer.set_resize_direction(resizers.ResizeDirection.Left | resizers.ResizeDirection.Top)
        self._top_right_resizer.set_resize_direction(resizers.ResizeDirection.Right | resizers.ResizeDirection.Top)
        self._bottom_left_resizer.set_resize_direction(resizers.ResizeDirection.Left | resizers.ResizeDirection.Bottom)
        self._bottom_right_resizer.set_resize_direction(
            resizers.ResizeDirection.Right | resizers.ResizeDirection.Bottom)

    def get_resizers_height(self):
        """
        Returns the total height of the vertical resizers
        :return: float
        """

        resizers = [self._top_resizer, self._bottom_resizer]
        total_height = 0
        for r in resizers:
            if not r.isHidden():
                total_height += r.minimumSize().height()

        return total_height

    def get_resizers_width(self):
        """
        Returns the total widht of the horizontal resizers
        :return: float
        """

        resizers = [self._left_resizer, self._right_resizer]
        total_width = 0
        for r in resizers:
            if not r.isHidden():
                total_width += r.minimumSize().width()

        return total_width

    def _setup_resizers(self):
        """
        Internal function that setup window resizers
        """

        self._top_resizer = resizers.VerticalResizer(parent=self)
        self._bottom_resizer = resizers.VerticalResizer(parent=self)
        self._right_resizer = resizers.HorizontalResizer(parent=self)
        self._left_resizer = resizers.HorizontalResizer(parent=self)
        self._top_left_resizer = resizers.CornerResizer(parent=self)
        self._top_right_resizer = resizers.CornerResizer(parent=self)
        self._bottom_left_resizer = resizers.CornerResizer(parent=self)
        self._bottom_right_resizer = resizers.CornerResizer(parent=self)

        self._resizers = [
            self._top_resizer, self._top_right_resizer, self._right_resizer, self._bottom_right_resizer,
            self._bottom_resizer, self._bottom_left_resizer, self._left_resizer, self._top_left_resizer
        ]

    def _get_title_pixmap(self):
        """
        Internal function that sets the pixmap used for the title
        """

        return None


class BaseColorDialog(BaseDialog, object):

    def_title = 'Select Color'

    maya_colors = [
        (.467, .467, .467), (.000, .000, .000), (.247, .247, .247), (.498, .498, .498), (0.608, 0, 0.157),
        (0, 0.016, 0.373), (0, 0, 1), (0, 0.275, 0.094), (0.145, 0, 0.263), (0.78, 0, 0.78), (0.537, 0.278, 0.2),
        (0.243, 0.133, 0.122), (0.6, 0.145, 0), (1, 0, 0), (0, 1, 0), (0, 0.255, 0.6), (1, 1, 1), (1, 1, 0),
        (0.388, 0.863, 1), (0.263, 1, 0.635), (1, 0.686, 0.686), (0.89, 0.675, 0.475), (1, 1, 0.384), (0, 0.6, 0.325),
        (0.627, 0.412, 0.188), (0.62, 0.627, 0.188), (0.408, 0.627, 0.188), (0.188, 0.627, 0.365),
        (0.188, 0.627, 0.627), (0.188, 0.404, 0.627), (0.435, 0.188, 0.627), (0.627, 0.188, 0.404)]

    def __init__(self, name='MayaColorDialog', parent=None, **kwargs):
        parent = parent or dcc.get_main_window()

        super(BaseColorDialog, self).__init__(name=name, parent=parent, **kwargs)

        self._color = None

    def get_color(self):
        return self._color

    color = property(get_color)

    def ui(self):

        self.color_buttons = list()

        super(BaseColorDialog, self).ui()

        grid_layout = layouts.GridLayout()
        grid_layout.setAlignment(Qt.AlignTop)
        self.main_layout.addLayout(grid_layout)
        color_index = 0
        for i in range(0, 4):
            for j in range(0, 8):
                color_btn = QPushButton()
                color_btn.setMinimumHeight(35)
                color_btn.setMinimumWidth(35)
                self.color_buttons.append(color_btn)
                color_btn.setStyleSheet('background-color:rgb(%s,%s,%s);' % (
                    self.maya_colors[color_index][0] * 255,
                    self.maya_colors[color_index][1] * 255,
                    self.maya_colors[color_index][2] * 255
                ))
                grid_layout.addWidget(color_btn, i, j)
                color_index += 1
        selected_color_layout = layouts.HorizontalLayout()
        self.main_layout.addLayout(selected_color_layout)
        self.color_slider = QSlider(Qt.Horizontal)
        self.color_slider.setMinimum(0)
        self.color_slider.setMaximum(31)
        self.color_slider.setValue(2)
        self.color_slider.setStyleSheet(
            "QSlider::groove:horizontal {border: 1px solid #999999;height: 25px; /* the groove expands "
            "to the size of the slider by default. by giving it a height, it has a fixed size */background: "
            "qlineargradient(x1:0, y1:0, x2:0, y2:1, stop:0 #B1B1B1, stop:1 #c4c4c4);margin: 2px 0;}"
            "QSlider::handle:horizontal {background: qlineargradient(x1:0, y1:0, x2:1, y2:1, stop:0 #b4b4b4,"
            " stop:1 #8f8f8f);border: 1px solid #5c5c5c;width: 10px;margin: -2px 0; /* handle is placed by "
            "default on the contents rect of the groove. Expand outside the groove */border-radius: 1px;}")
        selected_color_layout.addWidget(self.color_slider)

        color_label_layout = layouts.HorizontalLayout(margins=(10, 10, 10, 10))
        self.main_layout.addLayout(color_label_layout)

        self.color_lbl = QLabel()
        self.color_lbl.setStyleSheet("border: 1px solid black; background-color:rgb(0, 0, 0);")
        self.color_lbl.setMinimumWidth(45)
        self.color_lbl.setMaximumWidth(80)
        self.color_lbl.setMinimumHeight(80)
        self.color_lbl.setAlignment(Qt.AlignCenter)
        color_label_layout.addWidget(self.color_lbl)

        bottom_layout = layouts.HorizontalLayout()
        bottom_layout.setAlignment(Qt.AlignRight)
        self.main_layout.addLayout(bottom_layout)

        self.ok_btn = QPushButton('Ok')
        self.cancel_btn = QPushButton('Cancel')
        bottom_layout.addLayout(dividers.DividerLayout())
        bottom_layout.addWidget(self.ok_btn)
        bottom_layout.addWidget(self.cancel_btn)

    def setup_signals(self):

        for i, btn in enumerate(self.color_buttons):
            btn.clicked.connect(partial(self._on_set_color, i))
        self.color_slider.valueChanged.connect(self._on_set_color)

        self.ok_btn.clicked.connect(self._on_ok_btn)
        self.cancel_btn.clicked.connect(self._on_cancel_btn)

    def _on_set_color(self, color_index):
        self.color_lbl.setStyleSheet('background-color:rgb(%s,%s,%s);' % (
            self.maya_colors[color_index][0] * 255,
            self.maya_colors[color_index][1] * 255,
            self.maya_colors[color_index][2] * 255
        ))
        self.color_slider.setValue(color_index)

    def _on_set_slider(self, color_index):
        self._set_color(color_index=color_index)

    def _on_ok_btn(self):
        self._color = self.color_slider.value()
        self.close()

    def _on_cancel_btn(self):
        self._color = None
        self.close()


class BaseFileFolderDialog(BaseDialog, abstract_dialog.AbstractFileFolderDialog):
    """
    Base dialog classes for folders and files
    """

    def_title = 'Select File'
    def_size = (200, 125)
    def_use_app_browser = False

    def __init__(self,
                 name='BaseFileFolder', parent=None, **kwargs):
        super(BaseFileFolderDialog, self).__init__(name=name, parent=parent)

        self.directory = None
        self.filters = None
        self._use_app_browser = kwargs.pop('use_app_browser', self.def_use_app_browser)

        self.set_filters('All Files (*.*)')

        # By default, we set the directory to the user folder
        self.set_directory(os.path.expanduser('~'))
        self.center()

    def open_app_browser(self):
        return

    def ui(self):
        super(BaseFileFolderDialog, self).ui()

        from tpDcc.libs.qt.widgets import directory

        self.places = dict()

        self.grid = layouts.GridLayout()
        sub_grid = layouts.GridLayout()
        self.grid.addWidget(QLabel('Path:'), 0, 0, Qt.AlignRight)

        self.path_edit = QLineEdit(self)
        self.path_edit.setReadOnly(True)
        self.filter_box = QComboBox(self)
        self.file_edit = QLineEdit(self)

        self.view = directory.FileListWidget(self)
        self.view.setWrapping(True)
        self.view.setFocusPolicy(Qt.StrongFocus)

        self.open_button = QPushButton('Select', self)
        self.cancel_button = QPushButton('Cancel', self)

        size = QSize(32, 24)
        self.up_button = QPushButton('Up')
        self.up_button.setToolTip('Go up')
        self.up_button.setMinimumSize(size)
        self.up_button.setMaximumSize(size)

        size = QSize(56, 24)
        self.refresh_button = QPushButton('Reload')
        self.refresh_button.setToolTip('Reload file list')
        self.refresh_button.setMinimumSize(size)
        self.refresh_button.setMaximumSize(size)

        self.show_hidden = QCheckBox('Hidden')
        self.show_hidden.setChecked(False)
        self.show_hidden.setToolTip('Toggle show hidden files')

        sub_grid.addWidget(self.up_button, 0, 1)
        sub_grid.addWidget(self.path_edit, 0, 2)
        sub_grid.addWidget(self.refresh_button, 0, 3)
        sub_grid.addWidget(self.show_hidden, 0, 4)
        self.grid.addLayout(sub_grid, 0, 1)
        self.grid.addWidget(self.get_drives_widget(), 1, 0)
        self.grid.addWidget(self.view, 1, 1)
        self.grid.addWidget(QLabel('File name:'), 7, 0, Qt.AlignRight)
        self.grid.addWidget(self.file_edit, 7, 1)
        self.filter_label = QLabel('Filter:')
        self.grid.addWidget(self.filter_label, 8, 0, Qt.AlignRight)
        self.grid.addWidget(self.filter_box, 8, 1)
        hbox = layouts.GridLayout()
        hbox.addWidget(self.open_button, 0, 0, Qt.AlignRight)
        hbox.addWidget(self.cancel_button, 0, 1, Qt.AlignRight)
        self.grid.addLayout(hbox, 9, 1, Qt.AlignRight)
        self.main_layout.addLayout(self.grid)
        self.setGeometry(200, 100, 600, 400)

        self.open_button.clicked.connect(self.accept)
        self.cancel_button.clicked.connect(self.reject)
        self.up_button.clicked.connect(self.go_up)
        self.refresh_button.clicked.connect(self.update_view)
        self.show_hidden.stateChanged.connect(self.update_view)
        self.view.directory_activated.connect(self.activate_directory_from_view)
        self.view.file_activated.connect(self.activate_file_from_view)
        self.view.file_selected.connect(self.select_file_item)
        self.view.folder_selected.connect(self.select_folder_item)
        self.view.up_requested.connect(self.go_up)
        self.view.update_requested.connect(self.update_view)

    def exec_(self, *args, **kwargs):
        if self._use_app_browser:
            return self.open_app_browser()
        else:
            self.update_view()
            self.filter_box.currentIndexChanged.connect(self.update_view)
            accepted = super(BaseFileFolderDialog, self).exec_()
            self.filter_box.currentIndexChanged.disconnect(self.update_view)
            return self.get_result() if accepted == 1 else None

    def set_filters(self, filters, selected=0):
        self.filter_box.clear()
        filter_types = filters.split(';;')
        for ft in filter_types:
            extensions = string.extract(ft, '(', ')')
            filter_name = string.rstrips(ft, '({})'.format(extensions))
            extensions = extensions.split(' ')
            self.filter_box.addItem('{} ({})'.format(filter_name, ','.join(extensions)), extensions)
        if 0 <= selected < self.filter_box.count():
            self.filter_box.setCurrentIndex(selected)
        self.filters = filters

    def get_drives_widget(self):
        """
        Returns a QGroupBox widget that contains all disk drivers of the PC in a vertical layout
        :return: QGroupBox
        """

        w = QGroupBox('')
        w.setParent(self)
        box = layouts.VerticalLayout()
        box.setAlignment(Qt.AlignTop)
        places = [(getpass.getuser(), os.path.realpath(os.path.expanduser('~')))]
        places += [(q, q) for q in [os.path.realpath(x.absolutePath()) for x in QDir().drives()]]
        for label, loc in places:
            icon = QFileIconProvider().icon(QFileInfo(loc))
            drive_btn = QRadioButton(label)
            drive_btn.setIcon(icon)
            drive_btn.setToolTip(loc)
            drive_btn.setProperty('path', loc)
            drive_btn.clicked.connect(self.go_to_drive)
            self.places[loc] = drive_btn
            box.addWidget(drive_btn)
        w.setLayout(box)
        return w

    def go_to_drive(self):
        """
        Updates widget to show the content of the selected disk drive
        """

        sender = self.sender()
        self.set_directory(sender.property('path'), False)

    def get_result(self):
        tf = self.file_edit.text()
        sf = self.get_file_path(tf)
        return sf, os.path.dirname(sf), tf.split(os.pathsep)

    def get_filter_patterns(self):
        """
        Get list of filter patterns that are being used by the widget
        :return: list<str>
        """

        idx = self.filter_box.currentIndex()
        if idx >= 0:
            return self.filter_box.itemData(idx)
        else:
            return []

    def get_file_path(self, file_name):
        """
        Returns file path of the given file name taking account the selected directory
        :param file_name: str, name of the file without path
        :return: str
        """

        sname = file_name.split(os.pathsep)[0]
        return os.path.realpath(os.path.join(os.path.abspath(self.directory), sname))

#     def accept(self):
#         self._overlay.close()
#         super(BaseFileFolderDialog, self).accept()
#
#
#     def reject(self):
#         self._overlay.close()
#         super(BaseFileFolderDialog, self).reject()

    def update_view(self):
        """
        Updates file/folder view
        :return:
        """

        self.view.clear()
        qdir = QDir(self.directory)
        qdir.setNameFilters(self.get_filter_patterns())
        filters = QDir.Dirs | QDir.AllDirs | QDir.Files | QDir.NoDot | QDir.NoDotDot
        if self.show_hidden.isChecked():
            filters = filters | QDir.Hidden
        entries = qdir.entryInfoList(filters=filters, sort=QDir.DirsFirst | QDir.Name)
        file_path = self.get_file_path('..')
        if os.path.exists(file_path) and file_path != self.directory:
            icon = QFileIconProvider().icon(QFileInfo(self.directory))
            QListWidgetItem(icon, '..', self.view, 0)
        for info in entries:
            icon = QFileIconProvider().icon(info)
            suf = info.completeSuffix()
            name, tp = (
                info.fileName(), 0) if info.isDir() else ('%s%s' % (info.baseName(), '.%s' % suf if suf else ''), 1)
            QListWidgetItem(icon, name, self.view, tp)
        self.view.setFocus()

    def set_directory(self, path, check_drive=True):
        """
        Sets the directory that you want to explore
        :param path: str, valid path
        :param check_drive: bool,
        :return:
        """

        self.directory = os.path.realpath(path)
        self.path_edit.setText(self.directory)
        self.file_edit.setText('')

        # If necessary, update selected disk driver
        if check_drive:
            for loc in self.places:
                rb = self.places[loc]
                rb.setAutoExclusive(False)
                rb.setChecked(loc.lower() == self.directory.lower())
                rb.setAutoExclusive(True)

        self.update_view()
        self.up_button.setEnabled(not self.cant_go_up())

    def go_up(self):
        """
        Updates the current directory to go to its parent directory
        """

        self.set_directory(os.path.dirname(self.directory))

    def cant_go_up(self):
        """
        Checks whether we can naviage to current selected parent directory or not
        :return: bool
        """

        return os.path.dirname(self.directory) == self.directory

    def activate_directory_from_view(self, name):
        """
        Updates selected directory
        :param name: str, name of the directory
        """

        self.set_directory(os.path.join(self.directory, name))

    def activate_file_from_view(self, name):
        """
        Updates selected file text and returns its info by accepting it
        :param name: str, name of the file
        """

        self.select_file_item(name=name)
        self.accept()

    def select_file_item(self, name):
        """
        Updates selected file text and returns its info by accepting it
        :param name: str, name of the file
        """

        self.file_edit.setText(name)

    def select_folder_item(self, name):
        """
        Updates selected folder text and returns its info by accepting it
        :param name: str, name of the folder
        """

        self.file_edit.setText(name)


class BaseOpenFileDialog(BaseFileFolderDialog, object):
    """
    Open file dialog
    """

    def __init__(
            self,
            name='OpenFile',
            multi=False,
            title='Open File',
            size=(200, 125),
            fixed_size=False,
            frame_less=True,
            hide_title=False,
            parent=None,
            use_app_browser=False):

        parent = parent or dcc.get_main_window()

        super(BaseOpenFileDialog, self).__init__(
            name=name, title=title, size=size, fixed_size=fixed_size, frame_less=frame_less,
            hide_title=hide_title, use_app_browser=use_app_browser, parent=parent
        )

        self._multi = multi
        if multi:
            self.setExtendedSelection()

    def accept(self, *args, **kwargs):
        selected_file, selected_dir, selected_file_name = self.get_result()
        if not os.path.isdir(selected_file):
            if os.path.exists(selected_file):
                super(BaseOpenFileDialog, self).accept()
            else:
                message_box = QMessageBox()
                message_box.setWindowTitle('Confirme file selection')
                message_box.setText('File "{0}" does not exists!'.format(selected_file))
                message_box.exec_()

    def select_file_item(self, names):
        if self._multi:
            self.file_edit.setText(os.pathsep.join(names))
        else:
            super(BaseOpenFileDialog, self).select_file_item(names)


class BaseSaveFileDialog(BaseFileFolderDialog, object):
    def __init__(self,
                 name='SaveFile',
                 title='Save File',
                 size=(200, 125),
                 fixed_size=False,
                 frame_less=True,
                 hide_title=False,
                 parent=None,
                 use_app_browser=False):

        parent = parent or dcc.get_main_window()

        super(BaseSaveFileDialog, self).__init__(
            name=name, title=title, size=size, fixed_size=fixed_size, frame_less=frame_less, hide_title=hide_title,
            use_app_browser=use_app_browser, parent=parent)

        self._open_button.setText('Save')
        size = QSize(42, 24)
        self.new_directory_button = QPushButton('New')
        self.new_directory_button.setToolTip('Create new directory')
        self.new_directory_button.setMinimumSize(size)
        self.new_directory_button.setMaximumWidth(size)
        self.new_directory_button.clicked.connect(self.create_new_directory)
        self.grid.itemAtPosition(0, 1).addWidget(self.new_directory_button, 0, 5)

    def accept(self, *args, **kwargs):
        selected_file, selected_dir, selected_filename = self.get_result()
        if not os.path.isdir(selected_file):
            if os.path.exists(selected_file):
                message_box = QMessageBox()
                message_box.setWindowTitle('Confirm File Selection')
                message_box.setText('File "%s" exists.\nDo you want to overwrite it?' % selected_file)
                message_box.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
                message_box.setDefaultButton(QMessageBox.No)
                rv = message_box.exec_()
                if rv == QMessageBox.Yes and not os.path.isdir(selected_file):
                    super(BaseSaveFileDialog, self).accept()
        else:
            super(BaseSaveFileDialog, self).accept()

    def create_new_directory(self):
        name, ok = QInputDialog.getText(self, 'New directory name', 'Name:', QLineEdit.Normal, 'New Directory')
        if ok and name:
            path = os.path.join(self.directory, name)
            if os.path.exists(path):
                msg_box = QMessageBox(self)
                msg_box.setWindowTitle('Error')
                msg_box.setText('Directory already exists')
                msg_box.exec_()
            else:
                try:
                    os.makedirs(path)
                    self.update_view()
                except os.error as e:
                    msg_box = QMessageBox(self)
                    msg_box.setWindowTitle('Error')
                    msg_box.setText('Cannot create directory')
                    msg_box.exec_()


class BaseSelectFolderDialog(BaseFileFolderDialog, object):
    def __init__(self,
                 name='SelectFolder',
                 title='Select Folder',
                 size=(200, 125),
                 fixed_size=False,
                 frame_less=True,
                 hide_title=False,
                 use_app_browser=False,
                 parent=None,
                 **kwargs):

        parent = parent or dcc.get_main_window()

        super(BaseSelectFolderDialog, self).__init__(
            name=name, title=title, size=size, fixed_size=fixed_size, frame_less=frame_less, hide_title=hide_title,
            use_app_browser=use_app_browser, parent=parent, **kwargs
        )

    def accept(self, *args, **kwargs):
        selected_file, selected_dir, selected_filename = self.get_result()
        super(BaseSelectFolderDialog, self).accept()

    def exec_(self, *args, **kwargs):
        self.set_filters('')
        return super(BaseSelectFolderDialog, self).exec_()


class BaseNativeDialog(abstract_dialog.AbstractNativeDialog, object):
    """
    Dialog that opens DCC native dialogs
    """

    pass


class DialogContents(QFrame, object):
    """
    Widget that defines the core contents of frameless window
    Can be used to custom CSS for frameless windows contents
    """

    def __init__(self, parent=None):
        super(DialogContents, self).__init__(parent=parent)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains classes to create different types of lists
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, Signal, QRect, QPropertyAnimation, QEasingCurve, QSequentialAnimationGroup, QAbstractAnimation
from Qt.QtWidgets import QWidget, QFrame, QPushButton, QLineEdit, QListWidget, QTreeWidget, QListWidgetItem
from Qt.QtWidgets import QAbstractItemView

from tpDcc.libs.qt.widgets import layouts


class EditableList(QTreeWidget):
    """
    List list with editable list items
    """

    itemUpdated = Signal(str, str)
    editing = False

    class EditLine(QLineEdit):
        def __init__(self, *args):
            super(EditableList.EditLine, self).__init__(*args)

        def keyPressEvent(self, event):
            if event.key() == Qt.Key_Escape:
                self.deleteLater()
            elif event.key() in (Qt.Key_Return, Qt.Key_Enter):
                self.parent().parent().update_item()
            else:
                super(EditableList.EditLine, self).keyPressEvent(event)

    def __init__(self, parent=None):
        super(EditableList, self).__init__(parent=parent)
        self.itemClicked.connect(self._on_click)
        self.itemDoubleClicked.connect(self._on_double_click)

    def update_item(self):
        """
        If an item has been edited we update the list
        """

        i = self.editing
        w = self.itemWidget(i, 0)

        if w and not self.findItems(w.text(), Qt.MatchExactly, 0):
            self.itemUpdated.emit(i.text(0), w.text())
            i.setText(0, w.text())
            self.scrollToItem(i)
            self.editing = False
        if i:
            self.removeItemWidget(i, 0)

    def _on_click(self, item, col):
        if self.editing and item is not self.editing:
            self.update_item()

    def _on_double_click(self, item, col):
        edit = EditableList.EditLine(item.text(0), self)
        self.setItemWidget(item, 0, edit)
        edit.selectAll()
        self.editing = item


class WidgetItem(QFrame, object):

    closed = Signal(QWidget)
    deleted = Signal(QWidget)

    def __init__(self, item_height=150, item_width=300, height_offset=10, width_offset=10,
                 has_title=True, editable_title=True, is_closable=True, parent=None):
        super(WidgetItem, self).__init__(parent=parent)

        self._item_height = item_height
        self._item_width = item_width
        self._height_offset = height_offset
        self._width_offset = width_offset
        self._has_title = has_title
        self._editable_title = editable_title
        self._is_closable = is_closable
        self._animation = None

        self.setFrameStyle(QFrame.Panel | QFrame.Raised)

        self.ui()

    def ui(self):

        self.setLayout(layouts.VerticalLayout(spacing=0, margins=(3, 1, 3, 3)))

        self.main_layout = layouts.VerticalLayout(spacing=5, margins=(2, 2, 2, 2))
        main_widget = QWidget()

        if self._item_height:
            main_widget.setFixedHeight(self._item_height - self._height_offset)
        if self._item_width:
            main_widget.setFixedWidth(self._item_width - self._width_offset)

        main_widget.setLayout(self.main_layout)
        self.layout().addWidget(main_widget)

        # =====================================================

        # This layout is used to add custom widgets before the title of the node
        self.buttons_layout = layouts.VerticalLayout(spacing=2, margins=(2, 2, 2, 2))

        title_layout = layouts.HorizontalLayout()
        title_layout.addLayout(self.buttons_layout)

        title_line = QLineEdit('Untitled')
        if not self._editable_title:
            title_line.setEnabled(False)

        if self._has_title:
            title_layout.addWidget(title_line)

        self._close_btn = QPushButton('X')
        self._close_btn.setFixedHeight(20)
        self._close_btn.setFixedWidth(20)
        self._close_btn.clicked.connect(self.close_widget)
        if self._is_closable:
            title_layout.addWidget(self._close_btn)

        self.main_layout.addLayout(title_layout)

    def hide_close_button(self, value=True):
        self._close_btn.setVisible(not(value))

    def close_widget(self):
        if self._is_closable:
            self.closed.emit(self)

    def delete_widget(self):
        self.deleted.emit(self)

    def _animate_expand(self, value):

        size_anim = QPropertyAnimation(self, 'geometry')
        geometry = self.geometry()
        width = geometry.width()
        x, y, _, _ = geometry.getCoords()
        size_start = QRect(x, y, width, int(not(value)) * 150)
        size_end = QRect(x, y, width, value * 150)
        size_anim.setStartValue(size_start)
        size_anim.setEndValue(size_end)
        size_anim.setDuration(300)
        size_anim_curve = QEasingCurve()
        if value:
            size_anim_curve.setType(QEasingCurve.InQuad)
        else:
            size_anim_curve.setType(QEasingCurve.OutQuad)
        size_anim.setEasingCurve(size_anim_curve)

        # =================================================== Animation Sequence

        self._animation = QSequentialAnimationGroup()
        self._animation.addAnimation(size_anim)
        size_anim.valueChanged.connect(self._force_resize)
        if not value:
            self._animation.finished.connect(self.delete_widget)
        self._animation.start(QAbstractAnimation.DeleteWhenStopped)

    def _force_resize(self, new_height):

        # Force widget item parent to reevaluate its size
        self.setFixedHeight(new_height.height())


class ItemWidgetsList(QListWidget, object):
    def __init__(self, parent=None):
        super(ItemWidgetsList, self).__init__(parent=parent)

        self.setDragDropMode(QAbstractItemView.InternalMove)
        self.setDefaultDropAction(Qt.MoveAction)
        self.setFocusPolicy(Qt.NoFocus)

    def addWidget(self, widget):

        item = QListWidgetItem(self)
        item.setSizeHint(widget.sizeHint())
        super(ItemWidgetsList, self).addItem(item)


class WidgetsList(QWidget, object):

    def __init__(self, parent=None):
        super(WidgetsList, self).__init__(parent=parent)

        self._widgets_list = list()

        self.main_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        self.main_layout.setAlignment(Qt.AlignTop)
        self.setLayout(self.main_layout)

        self.widgets_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        self.main_layout.addLayout(self.widgets_layout)

        button_layout = layouts.HorizontalLayout(margins=(5, 5, 5, 5))
        button_layout.setAlignment(Qt.AlignRight)
        self.main_layout.addLayout(button_layout)

        self.add_button = QPushButton('New...', parent=self)
        button_layout.addWidget(self.add_button)

    def add_widget(self, widget):
        self.widgets_layout.addWidget(widget)
        self._widgets_list.append(widget)

        # The first widget of the list cannot be deleted
        if len(self._widgets_list) <= 1:
            self._widgets_list[0].hide_close_button(True)

        # When the close button is pressed we delete the widget
        widget.closed.connect(self.remove_widget)

        widget.setFixedHeight(0)
        widget._animate_expand(True)

    def remove_widget(self, widget):

        # We use try/catch to avoid an error when closing/creating new widgets too quicly in the UI
        widget.deleted.connect(self.delete_widget)
        self._widgets_list.remove(widget)
        widget._animate_expand(False)

    def delete_widget(self, widget):
        self.widgets_layout.removeWidget(widget)
        widget._animation = None
        widget.deleteLater()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains toast widget implementation
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, Signal, QPoint, QSize, QTimer, QPropertyAnimation, QEasingCurve, QAbstractAnimation
from Qt.QtWidgets import QLabel
from Qt.QtGui import QFontMetricsF

from tpDcc import dcc
from tpDcc.managers import resources
from tpDcc.libs.resources.core import theme
from tpDcc.libs.qt.core import base, animation
from tpDcc.libs.qt.widgets import layouts, label, avatar, loading


@theme.mixin
class BaseToast(base.BaseWidget, object):

    class ToastTypes(object):
        INFO = 'info'
        SUCCESS = 'success'
        WARNING = 'warning'
        ERROR = 'error'
        LOADING = 'loading'

    DEFAULT_CONFIG = {'duration': 2}

    toastClosed = Signal()

    def __init__(self, text, duration=None, toast_type=None, parent=None):
        self._text = text
        self._duration = duration
        self._toast_type = toast_type
        self._parent = parent
        super(BaseToast, self).__init__(parent=parent)

    def get_main_layout(self):
        main_layout = layouts.VerticalLayout(margins=(0, 0, 0, 0))

        return main_layout

    def ui(self):
        super(BaseToast, self).ui()

        self.setWindowFlags(Qt.FramelessWindowHint | Qt.Dialog | Qt.WA_TranslucentBackground | Qt.WA_DeleteOnClose)
        self.setAttribute(Qt.WA_StyledBackground)
        self.setFixedSize(QSize(120, 120))

        icon_layout = layouts.HorizontalLayout()
        icon_layout.addStretch()

        widget_theme = self.theme()

        if self._toast_type == self.ToastTypes.LOADING:
            icon_layout.addWidget(loading.CircleLoading(size=widget_theme.huge, color=widget_theme.text_color_inverse))
        else:
            icon_label = avatar.Avatar()
            icon_label.theme_size = 60
            icon_label.image = resources.pixmap(
                self._toast_type or self.ToastTypes.INFO, color=widget_theme.text_color_inverse)
            icon_layout.addWidget(icon_label)
        icon_layout.addStretch()

        content_label = label.BaseLabel()
        content_label.setText(self._text or '')
        content_label.setAlignment(Qt.AlignCenter)

        self.main_layout.addStretch()
        self.main_layout.addLayout(icon_layout)
        self.main_layout.addSpacing(10)
        self.main_layout.addWidget(content_label)
        self.main_layout.addStretch()

        close_timer = QTimer(self)
        close_timer.setSingleShot(True)
        close_timer.timeout.connect(self.close)
        close_timer.timeout.connect(self.toastClosed.emit)
        close_timer.setInterval((self._duration or self.DEFAULT_CONFIG.get('duration', 2)) * 1000)

        anim_timer = QTimer(self)
        anim_timer.timeout.connect(self._fade_out)
        anim_timer.setInterval((self._duration or self.DEFAULT_CONFIG.get('duration', 2)) * 1000 - 300)

        self._opacity_anim = QPropertyAnimation()
        self._opacity_anim.setTargetObject(self)
        self._opacity_anim.setDuration(300)
        self._opacity_anim.setEasingCurve(QEasingCurve.OutCubic)
        self._opacity_anim.setPropertyName('windowOpacity')
        self._opacity_anim.setStartValue(0.0)
        self._opacity_anim.setEndValue(0.9)

        close_timer.start()
        anim_timer.start()

        self._get_center_position(self._parent)
        self._fade_in()

    @classmethod
    def info(cls, text, parent, duration=None):
        inst = cls(text, duration=duration, toast_type=cls.ToastTypes.INFO, parent=parent)
        inst.show()

        return inst

    @classmethod
    def success(cls, text, parent, duration=None):
        inst = cls(text, duration=duration, toast_type=cls.ToastTypes.SUCCESS, parent=parent)
        inst.show()

        return inst

    @classmethod
    def warning(cls, text, parent, duration=None):
        inst = cls(text, duration=duration, toast_type=cls.ToastTypes.WARNING, parent=parent)
        inst.show()

        return inst

    @classmethod
    def error(cls, text, parent, duration=None):
        inst = cls(text, duration=duration, toast_type=cls.ToastTypes.ERROR, parent=parent)
        inst.show()

        return inst

    @classmethod
    def loading(cls, text, parent, duration=None):
        inst = cls(text, duration=duration, toast_type=cls.ToastTypes.LOADING, parent=parent)
        inst.show()

        return inst

    @classmethod
    def config(cls, duration):
        if duration is not None:
            cls.DEFAULT_CONFIG['duration'] = duration

    def _fade_out(self):
        self._opacity_anim.setDirection(QAbstractAnimation.Backward)
        self._opacity_anim.start()

    def _fade_in(self):
        self._opacity_anim.start()

    def _get_center_position(self, parent):
        parent_parent = parent.parent()
        dcc_win = dcc.get_main_window()
        if dcc_win:
            dcc_window = parent_parent == dcc_win or parent_parent.objectName() == dcc_win.objectName()
        else:
            dcc_window = None
        parent_geo = parent.geometry()
        pos = parent_geo.topLeft() if dcc_window else parent.mapToGlobal(parent_geo.topLeft())
        offset = 0
        for child in parent.children():
            if isinstance(child, BaseToast) and child.isVisible():
                offset = max(offset, child.y())
        target_x = pos.x() + parent_geo.width() / 2 - self.width() / 2
        target_y = pos.y() + parent_geo.height() / 2 - self.height() / 2
        self.setProperty('pos', QPoint(target_x, target_y))


class ToastWidget(QLabel, object):
    """
    Toast widget used to show quick messages to user
    """

    DEFAULT_DURATION = 500
    DEFAULT_PADDING = 30

    def __init__(self, *args):
        super(ToastWidget, self).__init__(*args)

        self._timer = QTimer(self)
        self._timer.setSingleShot(True)
        self._timer.timeout.connect(self._on_fade_out)

        self._duration = self.DEFAULT_DURATION

        self.setMouseTracking(True)
        self.setAlignment(Qt.AlignCenter)
        self.setAttribute(Qt.WA_TransparentForMouseEvents)

        if self.parent():
            self.parent().installEventFilter(self)

    # def eventFilter(self, obj, event):
    #     """
    #     Overrides base QLabel eventFilter function
    #     Updates the geometry when the parent widget changes size
    #     :param obj: QWidget
    #     :param event: QEvent
    #     """
    #
    #     if event.type() == QEvent.Resize:
    #         self.updateGeometry()
    #     return super(ToastWidget, self).eventFilter(obj, event)

    def updateGeometry(self):
        """
        Overrides base QLabel updateGeometry function
        Updates and aligns the geometry to the parent widget
        """

        padding = self.DEFAULT_PADDING
        widget = self.parent()

        width = self.text_width() + padding
        height = self.text_height() + padding
        x = widget.width() * 0.5 - width * 0.5
        y = (widget.height() - height) / 1.2

        self.setGeometry(x, y, width, height)

    def setText(self, *args, **kwargs):
        """
        Overrides base QLabel setText function
        Updates the size depending on the text width
        :param text: str
        """

        super(ToastWidget, self).setText(*args, **kwargs)
        self.updateGeometry()

    def show(self):
        """
        Overrides base QLabel show function
        Starts the timer to hide the toast
        """

        duration = self.duration()
        self._timer.stop()
        self._timer.start(duration)
        if not self.isVisible():
            animation.fade_in_widget(self, duration=0)
            super(ToastWidget, self).show()

    def duration(self):
        """
        Returns duration
        :return: int
        """

        return self._duration

    def set_duration(self, duration):
        """
        Sets how long to show the toast (in milliseconds)
        :param duration: int
        """

        self._duration = duration

    def text_rect(self):
        """
        Returns the bounding box rect for the text
        :return: QRect
        """

        text = self.text()
        font = self.font()
        metrics = QFontMetricsF(font)

        return metrics.boundingRect(text)

    def text_width(self):
        """
        Returns the width of the text
        :return: int
        """

        text_width = self.text_rect().width()
        return max(0, text_width)

    def text_height(self):
        """
        Returns the height of the text
        :return: int
        """

        text_height = self.text_rect().height()
        return max(0, text_height)

    def _on_fade_out(self, duration=250):
        """
        Internal callback function that fades out the toast message
        :param duration: int
        """

        animation.fade_out_widget(self, duration=duration, on_finished=self.hide)



import tpDcc
from tpDcc.libs.qt.widgets import layouts, label, loading, dividers


class LoadingExample(tpDcc.Window, object):
    def __init__(self):
        super(LoadingExample, self).__init__(title='Example for Loading Widgets')

    def ui(self):
        super(LoadingExample, self).ui()

        size_lyt = layouts.HorizontalLayout()
        size_list = [
            ('Tiny', loading.CircleLoading.tiny),
            ('Small', loading.CircleLoading.small),
            ('Medium', loading.CircleLoading.medium),
            ('Large', loading.CircleLoading.large),
            ('Huge', loading.CircleLoading.huge)
        ]
        for lbl, load_widget in size_list:
            size_lyt.addWidget(label.BaseLabel(lbl))
            size_lyt.addWidget(load_widget(parent=self))
            size_lyt.addSpacing(10)

        color_lyt = layouts.HorizontalLayout()
        color_list = [
            ('cyan', '#13c2c2'),
            ('green', '#52c41a'),
            ('magenta', '#eb2f96'),
            ('red', '#f5222d'),
            ('yellow', '#fadb14'),
            ('volcano', '#fa541c'),
        ]
        for lbl, load_color in color_list:
            color_lyt.addWidget(label.BaseLabel(lbl))
            color_lyt.addWidget(loading.CircleLoading(color=load_color, parent=self))
            color_lyt.addSpacing(10)

        self.main_layout.addWidget(dividers.Divider('different size'))
        self.main_layout.addLayout(size_lyt)
        self.main_layout.addWidget(dividers.Divider('different color'))
        self.main_layout.addLayout(color_lyt)


def run():
    return LoadingExample()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains classes to create attribute editors
"""

from __future__ import print_function, division, absolute_import

import logging

from Qt.QtCore import Qt, Signal, QPoint, QPointF, QSize, QRegExp
from Qt.QtWidgets import QWidget, QGroupBox, QScrollArea, QLineEdit, QCheckBox, QSlider, QColorDialog
from Qt.QtGui import QColor, QPalette

from tpDcc.libs.python import strings as string_utils
from tpDcc.libs.qt.core import qtutils, base, color
from tpDcc.libs.qt.widgets import layouts, lineedit, directory

LOGGER = logging.getLogger('tpDcc-libs-qt')


class AttributeEditor(base.BaseWidget, object):
    def __init__(self, name='AttributeEditor', label='Attribute Editor', parent=None):
        self._label = label
        self._objects = list()
        super(AttributeEditor, self).__init__(parent=parent)
        self.setObjectName(name)

    def get_objects(self):
        return self._objects

    objects = property(get_objects)

    def ui(self):
        super(AttributeEditor, self).ui()

        self._main_group = QGroupBox()
        self._main_group.setProperty('class', 'attr_main_group')
        self._main_group.setFlat(False)
        self._main_group.setTitle('')

        self._main_group_layout = self.get_attributes_layout()
        self._main_group.setLayout(self._main_group_layout)

        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)
        scroll_area.setStyleSheet('QScrollArea { background-color: rgb(57,57,57);}')
        scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        scroll_area.setWidget(self._main_group)

        self.main_layout.addWidget(scroll_area)

        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._on_create_context_menu)

    def get_attributes_layout(self):
        """
        Returns layout used for the attributes
        :return: QLayout
        """

        attributes_layout = layouts.HorizontalLayout(margins=(5, 5, 5, 5))
        attributes_layout.setAlignment(Qt.AlignTop)

        return attributes_layout

    def set_title(self, new_title):
        self._main_group.setTitle(new_title)

    def clear_layout(self, reset_title=True):
        """
        Clears main layout and resetes editor title
        """

        qtutils.clear_layout(self._main_group_layout)
        if reset_title:
            self._main_group.setTitle('')

    def _build_layout(self):
        pass

    def _on_create_context_menu(self):
        pass


class BaseEditor(QWidget, object):
    attr_type = 'None'
    valueChanged = Signal(object)

    def __init__(self, parent=None, **kwargs):
        super(BaseEditor, self).__init__(parent=parent)

        self._default_value = 0.0
        self._current_value = None

        self.main_layout = layouts.HorizontalLayout(spacing=3, margins=(1, 1, 1, 1), parent=self)

    def get_default_value(self):
        return self._default_value

    def set_default_value(self, value):
        if value != self._default_value:
            self._default_value = value

    def get_value(self):
        return None

    def set_value(self, value):
        pass

    default_value = property(get_default_value, set_default_value)
    value = property(get_value)

    def OnValueUpdated(self):
        if self.value != self._current_value:
            self._current_value = self.value
            self.valueChanged.emit(self)

    def initialize_editor(self):
        pass


class FloatEditor(BaseEditor, object):

    attr_type = 'float'

    def __init__(self, parent=None, **kwargs):
        super(FloatEditor, self).__init__(parent=parent, **kwargs)

        self._default_value = 0.0

        self.value_line = lineedit.FloatLineEdit(self)
        self.main_layout.addWidget(self.value_line)

    def get_value(self):
        return self.value_line.value

    value = property(get_value)

    def initialize_editor(self):
        editor_value = self.default_value

        node_values = self.values
        if node_values:
            if len(node_values) > 1:
                pass
            elif len(node_values) == 1:
                if node_values[0]:
                    editor_value = node_values[0]
                if node_values[0] is None:
                    editor_value = self.default_value

                self.value_line.blockSignals(True)
                self.value_line.setText(str(editor_value))
                self.value_line.blockSignals(False)
                self.value_line.valueChanged.connect(self.OnValueUpdated)

    def set_connected(self, conn):
        if conn != self._connection:
            self._connection = conn
            self.value_line.setText(conn)
            self.value_line.setEnabled(False)
            self.value_line.setProperty('class', 'Connected')


class Float2Editor(BaseEditor, object):

    attr_type = 'float'

    def __init__(self, parent=None, **kwargs):
        super(Float2Editor, self).__init__(parent=parent, **kwargs)

        self._default_value = (0.0, 0.0)

        self.value1_line = lineedit.FloatLineEdit(self)
        self.value2_line = lineedit.FloatLineEdit(self)
        self.main_layout.addWidget(self.value1_line)
        self.main_layout.addWidget(self.value2_line)

    def get_value(self):
        return self.value1_line.value, self.value2_line.value

    value = property(get_value)

    def initialize_editor(self):
        editor_value = self.default_value

        node_values = self.values
        if node_values:
            if len(node_values) > 1:
                pass
            elif len(node_values) == 1:
                if node_values[0]:
                    editor_value = node_values[0]

                self.value1_line.blockSignals(True)
                self.value2_line.blockSignals(True)

                if type(editor_value) in [QPoint, QPointF]:
                    self.value1_line.setText(str(editor_value.x()))
                    self.value2_line.setText(str(editor_value.y()))
                elif type(editor_value) in [list, tuple]:
                    self.value1_line.setText(str(editor_value[0]))
                    self.value2_line.setText(str(editor_value[1]))
                self.value1_line.blockSignals(False)
                self.value2_line.blockSignals(False)
                self.value1_line.valueChanged.connect(self.OnValueUpdated)
                self.value2_line.valueChanged.connect(self.OnValueUpdated)

    def set_connected(self, conn):
        if conn != self._connection:
            self._conection = conn
            self.value1_line.setText(conn)
            self.value1_line.setEnabled(False)
            self.value2_line.setEnabled(False)
            self.value1_line.setProperty('class', 'Connected')
            self.value2_line.setProperty('class', 'Connected')


class Float3Editor(BaseEditor, object):

    attr_type = 'float'

    def __init__(self, parent=None, **kwargs):
        super(Float3Editor, self).__init__(parent=parent, **kwargs)

        self._default_value = (0.0, 0.0, 0.0)

        self.value1_line = lineedit.FloatLineEdit(self)
        self.value2_line = lineedit.FloatLineEdit(self)
        self.value3_line = lineedit.FloatLineEdit(self)
        self.main_layout.addWidget(self.value1_line)
        self.main_layout.addWidget(self.value2_line)
        self.main_layout.addWidget(self.value3_line)

    def get_value(self):
        return self.value1_line.value(), self.value2_line.value(), self.value3_line.value()

    value = property(get_value)

    def initialize_editor(self):
        editor_value = self.default_value

        node_values = self.values
        if node_values:
            if len(node_values) > 1:
                pass
            elif len(node_values) == 1:
                if node_values[0]:
                    editor_value = node_values[0]

                self.value1_line.blockSignals(True)
                self.value2_line.blockSignals(True)
                self.value3_line.blockSignals(True)
                self.value1_line.setText(str(editor_value[0]))
                self.value2_line.setText(str(editor_value[1]))
                self.value3_line.setText(str(editor_value[2]))
                self.value1_line.blockSignals(False)
                self.value2line.blockSignals(False)
                self.value3line.blockSignals(False)
                self.value1_line.valueChanged.connect(self.OnValueUpdated)
                self.value2_line.valueChanged.connect(self.OnValueUpdated)
                self.value3_line.valueChanged.connect(self.OnValueUpdated)

    def set_connected(self, conn):
        if conn != self._connection:
            self._conection = conn
            self.value1_line.setText(conn)
            self.value1_line.setEnabled(False)
            self.value2_line.setEnabled(False)
            self.value3_line.setEnabled(False)
            self.value1_line.setProperty('class', 'Connected')
            self.value2_line.setProperty('class', 'Connected')
            self.value3_line.setProperty('class', 'Connected')


class StringEditor(BaseEditor, object):

    attr_type = 'str'

    def __init__(self, parent=None, **kwargs):
        super(StringEditor, self).__init__(parent=parent, **kwargs)

        self._default_value = ""
        self._clean_value = kwargs.get('clean', True)

        self.value_line = QLineEdit(self)
        reg_exp = QRegExp('^([a-zA-Z0-9_]+)')
        self.main_layout.addWidget(self.value_line)

        self.value_line.textEdited.connect(self._validate_text)
        self.value_line.editingFinished.connect(self.OnValueUpdated)
        self.value_line.returnPressed.connect(self.OnValueUpdated)

    def get_value(self):
        return str(self.value_line.text())

    value = property(get_value)

    def initialize_editor(self):
        editor_value = self.default_value

        node_values = self.values
        if node_values:
            if len(node_values) > 1:
                pass
            elif len(node_values) == 1:
                if node_values[0]:
                    editor_value = node_values[0]

                self.value_line.blockSignals(True)
                self.value_line.setText(str(editor_value))
                self.value_line.blockSignals(False)

    def set_connected(self, conn):
        if conn != self._connection:
            self._connection = conn
            self.value_line.setText(conn)
            self.value_line.setEnabled(False)
            self.value_line.setProperty('class', 'Connected')

    def _validate_text(self, text):
        """
        Validates the given value and update the current text
        :param text: str, text to validate
        """

        current_text = self.value
        if self._clean_value:
            cursor_pos = self.value_line.cursorPosition()
            cleaned = string_utils.clean_string(text=text)
            self.value_line.blockSignals(True)
            self.value_line.setText(cleaned)
            self.value_line.blockSignals(False)
            self.value_line.setCursorPosition(cursor_pos)


class BoolEditor(BaseEditor, object):

    attr_type = 'bool'

    def __init__(self, parent=None, **kwargs):
        super(BoolEditor, self).__init__(parent=parent, **kwargs)

        self._default_value = False

        self.cbx = QCheckBox(self)
        self.main_layout.addWidget(self.cbx)

        self.cbx.toggled.connect(self.OnValueUpdated)

    def get_value(self):
        return self.cbx.isChecked()

    def initialize_editor(self):
        editor_value = self.default_value

        node_values = self.values
        if node_values:
            if len(node_values) > 1:
                pass
            elif len(node_values) == 1:
                if node_values[0]:
                    editor_value = node_values[0]

                self.cbx.blockSignals(True)
                self.cbx.setChecked(editor_value)
                self.cbx.blockSignals(False)

    def set_connected(self, conn):
        if conn != self._connection:
            self._connection = conn
            self.value_line.setText(conn)
            self.cbx.setEnabled(False)
            self.value_line.setProperty('class', 'Connected')


class ColorPicker(BaseEditor, object):

    attr_type = 'color'

    def __init__(self, parent=None, **kwargs):
        super(ColorPicker, self).__init__(parent=parent, **kwargs)

        self._default_value = (125, 125, 125)
        self.attr = None

        self.normalized = kwargs.get('normalized', False)
        self.min = kwargs.get('min', 0)
        self.max = kwargs.get('max', 99)
        self.color = kwargs.get('color', QColor(1.0, 1.0, 1.0))
        self.mult = kwargs.get('mult', 0.1)

        self.color_swatch = color.ColorSwatch(parent=self, color=self.color, normalized=self.normalized)
        self.color_swatch.setMaximumSize(QSize(75, 20))
        self.color_swatch.setMinimumSize(QSize(75, 20))
        self.color_swatch.set_color(color=self.color)
        self.main_layout.addWidget(self.color_swatch)

        self.slider = QSlider(self)
        self.slider.setOrientation(Qt.Horizontal)
        self.slider.setValue(self.max)
        self.main_layout.addWidget(self.slider)

        self.set_max(self.max)
        self.set_min(self.min)

        self.slider.valueChanged.connect(self.OnSliderChanged)
        self.slider.sliderReleased.connect(self.OnSliderReleased)
        self.color_swatch.clicked.connect(self.OnColorPicked)

    # region Properties
    def get_rgb(self):
        return self.color_swatch.qcolor.getRgb()[0:3]

    def get_rgbF(self):
        return self.color_swatch.qcolor.getRgbF()[0:3]

    def get_hsv(self):
        return self.color_swatch.qcolor.getHsv()[0:3]

    def get_hsvF(self):
        return self.color_swatch.qcolor.getHsvF()[0:3]

    rgb = property(get_rgb)
    rgbF = property(get_rgbF)
    hsv = property(get_hsv)
    gsvF = property(get_hsvF)

    def OnSliderChanged(self):
        slider_value = float(self.slider.value())
        if not self._current_value:
            LOGGER.debug(
                'Caching color: (%d, %d, %d)' % (
                    self.color_swatch.color[0], self.color_swatch.color[1], self.color_swatch.color[2]))
            self._current_value = self.color_swatch.color
        current_color = QColor(*self._current_value)
        darker = 200 - slider_value
        new_color = current_color.darker(darker)
        self.color_swatch.qcolor = new_color
        self.color_swatch._update()

    def OnSliderReleased(self):
        """
        Updates items color when the slider handle is released
        """

        color = self.color_swatch.color
        self.color_swatch._update()
        self.OnValueUpdated()

    def OnColorPicked(self):
        """
        Event to call color picker
        """

        dialog = QColorDialog(self.color_swatch.qcolor, self)
        if dialog.exec_():
            self.color_swatch.setPalette(QPalette(dialog.currentColor()))
            self.color_swatch.qcolor = dialog.currentColor()
            self.color_swatch._update()
            self.OnValueUpdated()

    def get_value(self):
        return self.color_swatch.color

    def initialize_editor(self):
        editor_value = self.default_value

        node_values = self.values
        if node_values:
            if len(node_values) > 1:
                pass
            elif len(node_values) == 1:
                if node_values[0]:
                    editor_value = node_values[0]

                self.color_swatch.set_color(editor_value)

    def set_connected(self, conn):
        if conn != self._connection:
            self._connection = conn
            self.value_line.setText(conn)
            self.color_swatch.setEnabled(False)

    def sizeHint(self):
        return QSize(350, 27)

    def set_attr(self, value):
        self.attr = value
        return self.attr

    def get_attr(self):
        return self.attr

    def set_min(self, value):
        """
        Sets the slider minimum value
        :param value: int, slider minimum value
        """

        self.min = value
        self.slider.setMinimum(value)

    def set_max(self, value):
        """
        Sets the silder maximum value
        :param value: int, slider maximum value
        """

        self.max = value
        self.slider.setMaximum(value)

    def get_qcolor(self):
        return self.color_swatch.qcolor

    def set_color(self, value):
        return self.color_swatch.set_color(color=value)

    def _update(self):
        return self.color_swatch._update()


class FileEditor(BaseEditor, object):

    attr_type = 'file'

    def __init__(self, parent=None, **kwargs):
        super(FileEditor, self).__init__(parent=parent, **kwargs)

        line_widget = directory.SelectFile()
        self.main_layout.addWidget(line_widget)


# ===============================================================================

WIDGET_MAPPER = dict(
    float=[FloatEditor, 0.0],
    float2=[Float2Editor, [0.0, 0.0]],
    float3=[Float3Editor, [0.0, 0.0, 0.0]],
    bool=[BoolEditor, False],
    str=[StringEditor, ""],
    string=[StringEditor, ""],
    file=[FileEditor, ""],
    dir=[FileEditor, ""],
    # int=[IntEditor, 0],
    # int2=[Int2Editor, [0,0]],
    # int3=[Int3Editor, [0,0,0]],
    # int8=[IntEditor, 0],
    color=[ColorPicker, [172, 172, 172, 255]],
    # short2=[Float2Editor, [0.0, 0.0]],
    # node=[StringEditor, ""],
    # multi=[StringEditor, ""],
    # doc=[DocumentEditor, ""]
)


def map_widget(attr_type, name, parent=None):
    """
    Map the widget to the attribute type
    :param attr_type:
    :param name:
    :param parent:
    :return:
    """

    widget_type = attr_type.replace(' ', '').lower()
    if widget_type not in WIDGET_MAPPER:
        LOGGER.warning('Invalid Editor Class: "{0}" ("{1}")'.format(attr_type, name))
    else:
        cls = WIDGET_MAPPER.get(widget_type)[0]
        return cls(parent=parent, name=name)
    return



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains decorators for Qt
"""

from __future__ import print_function, division, absolute_import

from functools import wraps

from Qt.QtCore import Qt
from Qt.QtWidgets import QApplication
from Qt.QtGui import QCursor


def show_wait_cursor(fn):
    """
    Decorator that shows wait cursor during function execution
    :param fn:
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        cursor = QCursor(Qt.WaitCursor)
        QApplication.setOverrideCursor(cursor)
        try:
            return fn(*args, **kwargs)
        finally:
            QApplication.restoreOverrideCursor()

    return wrapper


def show_arrow_cursor(fn):
    """
    Decorator that shows arrow cursor during function execution
    :param fn:
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        cursor = QCursor(Qt.ArrowCursor)
        QApplication.setOverrideCursor(cursor)
        try:
            return fn(*args, **kwargs)
        finally:
            QApplication.restoreOverrideCursor()

    return wrapper
